"""Tests predict_activity module.
"""

import unittest

import numpy as np
import os

from adapt import alignment
from adapt.utils import predict_activity
from adapt.utils.version import get_project_path, get_latest_model_version

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestPredictor(unittest.TestCase):
    """Tests methods in the Predictor class.
    """

    def setUp(self):
        # Use the provided models with default thresholds
        dir_path = get_project_path()
        cla_path_all = os.path.join(dir_path, 'models', 'classify', 'cas13a')
        reg_path_all = os.path.join(dir_path, 'models', 'regress', 'cas13a')
        cla_version = get_latest_model_version(cla_path_all)
        reg_version = get_latest_model_version(reg_path_all)
        cla_path = os.path.join(cla_path_all, cla_version)
        reg_path = os.path.join(reg_path_all, reg_version)

        self.predictor = predict_activity.Predictor(cla_path, reg_path)

    def test_model_input_from_nt(self):
        # Make context (both ends) be all 'A'
        context_1 = 'A' * self.predictor.context_nt
        # Make guide and target both be the same, an arbitrary sequence
        guide_1 = 'ATCG'
        target_with_context_1 = context_1 + guide_1 + context_1

        # Make context (both ends) be all 'T'
        context_2 = 'T' * self.predictor.context_nt
        # Make guide and target both be the same, an arbitrary sequence
        guide_2 = 'CCGG'
        target_with_context_2 = context_2 + guide_2 + context_2

        pairs = [(target_with_context_1, guide_1),
                (target_with_context_2, guide_2)]

        expected_onehot_1 = [[1,0,0,0,0,0,0,0]]*self.predictor.context_nt
        expected_onehot_1 += [[1,0,0,0,1,0,0,0], [0,0,0,1,0,0,0,1],
            [0,1,0,0,0,1,0,0], [0,0,1,0,0,0,1,0]]
        expected_onehot_1 += [[1,0,0,0,0,0,0,0]]*self.predictor.context_nt

        expected_onehot_2 = [[0,0,0,1,0,0,0,0]]*self.predictor.context_nt
        expected_onehot_2 += [[0,1,0,0,0,1,0,0], [0,1,0,0,0,1,0,0],
            [0,0,1,0,0,0,1,0], [0,0,1,0,0,0,1,0]]
        expected_onehot_2 += [[0,0,0,1,0,0,0,0]]*self.predictor.context_nt

        expected = np.array([expected_onehot_1, expected_onehot_2])

        self.assertTrue(
                np.array_equal(self.predictor._model_input_from_nt(pairs),
                    expected))

    def test_classify_and_decide(self):
        # Target context is all A and, at guide, is all G
        # Guide is GC repeated, so has 14 mismatches against the target
        # Should be inactive
        target_with_context_1 = ('A'*self.predictor.context_nt +
                'G'*28 + 'A'*self.predictor.context_nt)
        guide_1 = 'GC'*14

        # Target context is all A and, at guide, is all G
        # Guide is all G, so matches target
        # Should be active
        target_with_context_2 = ('A'*self.predictor.context_nt +
                'G'*28 + 'A'*self.predictor.context_nt)
        guide_2 = 'G'*28

        pairs = [(target_with_context_1, guide_1), (target_with_context_2,
            guide_2)]
        pairs_onehot = self.predictor._model_input_from_nt(pairs)

        self.assertListEqual(self.predictor._classify_and_decide(pairs_onehot),
                [False, True])

    def test_regress(self):
        # Target context is all A (except G PFS) and, at guide, is all G
        # Guide is all G except all A in the seed region against the target
        # Should not be highly active
        target_with_context_1 = ('A'*self.predictor.context_nt +
                'G'*28 + 'G' + 'A'*(self.predictor.context_nt - 1))
        guide_1 = 'G'*9 + 'A'*9 + 'G'*10

        # Target context is all A and, at guide, is all G
        # Guide is all A, so matches target
        # Should (probably) be highly active
        target_with_context_2 = ('A'*self.predictor.context_nt +
                'A'*28 + 'A'*self.predictor.context_nt)
        guide_2 = 'A'*28

        pairs = [(target_with_context_1, guide_1), (target_with_context_2,
            guide_2)]
        pairs_onehot = self.predictor._model_input_from_nt(pairs)

        r = self.predictor._regress(pairs_onehot)
        self.assertLess(r[0], self.predictor.regression_threshold)
        self.assertGreater(r[1], self.predictor.regression_threshold)

    def test_determine_highly_active_computations(self):
        # Target context is all A and, at guide, is all G
        # Guide is GC repeated, so has 14 mismatches against the target
        # Should be inactive
        target_with_context_1 = ('A'*self.predictor.context_nt +
                'G'*28 + 'A'*self.predictor.context_nt)
        guide_1 = 'GC'*14

        # Target context is all A and, at guide, is all G
        # Guide is all A, so matches target
        # Should (probably) be highly active
        target_with_context_2 = ('A'*self.predictor.context_nt +
                'A'*28 + 'A'*self.predictor.context_nt)
        guide_2 = 'A'*28

        # Target context is all A (except G PFS) and, at guide, is all G
        # Guide is all G except all A in the seed region against the target
        # Should be inactive
        target_with_context_3 = ('A'*self.predictor.context_nt +
                'G'*28 + 'G' + 'A'*(self.predictor.context_nt - 1))
        guide_3 = 'G'*9 + 'A'*9 + 'G'*10

        # Target context is all A (except G PFS) and, at guide, is all G
        # Guide is all G except all A in the seed region against the target
        # Should be inactive
        target_with_context_4 = ('A'*self.predictor.context_nt +
                'G'*28 + 'G' + 'A'*(self.predictor.context_nt - 1))
        guide_4 = 'G'*9 + 'A'*9 + 'G'*10

        # Target context is all A and, at guide, is all G
        # Guide is all A, so matches target
        # Should (probably) be highly active
        target_with_context_5 = ('A'*self.predictor.context_nt +
                'A'*28 + 'A'*self.predictor.context_nt)
        guide_5 = 'A'*28

        pairs = [(target_with_context_1, guide_1),
                (target_with_context_2, guide_2),
                (target_with_context_3, guide_3),
                (target_with_context_4, guide_4),
                (target_with_context_5, guide_5)]

        # Use _run_models_and_memoize() to compute results, and use a
        # nonsense start position (-1) (it doesn't matter), and pull
        # output determinations about highly active
        self.predictor._run_models_and_memoize(-1, pairs)
        r = [self.predictor._memoized_evaluations[-1][pair] for pair in pairs]
        highly_active = [x[1] for x in r]

        self.assertListEqual(highly_active, [False, True, False, False, True])

    def test_memoized_evaluations(self):
        # Target context is all A and, at guide, is all G
        # Guide is GC repeated, so has 14 mismatches against the target
        # Should be inactive
        target_with_context_1 = ('A'*self.predictor.context_nt +
                'G'*28 + 'A'*self.predictor.context_nt)
        guide_1 = 'GC'*14

        # Target context is all A and, at guide, is all G
        # Guide is all A, so matches target
        # Should (probably) be highly active
        target_with_context_2 = ('A'*self.predictor.context_nt +
                'A'*28 + 'A'*self.predictor.context_nt)
        guide_2 = 'A'*28

        # Target context is all A (except G PFS) and, at guide, is all G
        # Guide is all G except all A in the seed region against the target
        # Should be inactive
        target_with_context_3 = ('A'*self.predictor.context_nt +
                'G'*28 + 'G' + 'A'*(self.predictor.context_nt - 1))
        guide_3 = 'G'*9 + 'A'*9 + 'G'*10

        pairs_1 = [(target_with_context_1, guide_1)]
        pairs_23 = [(target_with_context_2, guide_2),
                (target_with_context_3, guide_3)]

        self.assertListEqual(self.predictor.determine_highly_active(1, pairs_1),
                [False])
        self.assertListEqual(self.predictor.determine_highly_active(5, pairs_23),
                [True, False])
        self.assertEqual(self.predictor.compute_activity(1, pairs_1)[0], 0)
        self.assertGreater(self.predictor.compute_activity(5, pairs_23)[0], 0)
        self.assertEqual(self.predictor.compute_activity(5, pairs_23)[1], 0)
        self.predictor.cleanup_memoized(1)
        self.assertEqual(self.predictor.compute_activity(1, pairs_1)[0], 0)
        self.predictor.cleanup_memoized(1)

        # Target context is all A (except G PFS) and, at guide, is all G
        # Guide is all G except all A in the seed region against the target
        # Should be inactive
        target_with_context_4 = ('A'*self.predictor.context_nt +
                'G'*28 + 'G' + 'A'*(self.predictor.context_nt - 1))
        guide_4 = 'G'*9 + 'A'*9 + 'G'*10

        # Target context is all A and, at guide, is all G
        # Guide is all A, so matches target
        # Should (probably) be highly active
        target_with_context_5 = ('A'*self.predictor.context_nt +
                'A'*28 + 'A'*self.predictor.context_nt)
        guide_5 = 'A'*28

        pairs_45 = [(target_with_context_4, guide_4),
                (target_with_context_5, guide_5)]
        pairs_54 = [(target_with_context_5, guide_5),
                (target_with_context_4, guide_4)]

        # Should be memoized
        self.assertListEqual(self.predictor.determine_highly_active(5, pairs_45),
                [False, True])
        self.assertEqual(self.predictor.compute_activity(5, pairs_45)[0], 0)
        self.assertGreater(self.predictor.compute_activity(5, pairs_45)[1], 0)

        # Cleanup memoizations and test
        self.predictor.cleanup_memoized(5)
        self.assertListEqual(self.predictor.determine_highly_active(5, pairs_45),
                [False, True])
        self.assertListEqual(self.predictor.determine_highly_active(5, pairs_54),
                [True, False])
        self.assertEqual(self.predictor.compute_activity(5, pairs_45)[0], 0)
        self.assertGreater(self.predictor.compute_activity(5, pairs_45)[1], 0)
        self.assertGreater(self.predictor.compute_activity(5, pairs_54)[0], 0)
        self.assertEqual(self.predictor.compute_activity(5, pairs_54)[1], 0)

        # Do this again, with activity values first
        self.predictor.cleanup_memoized(5)
        self.assertEqual(self.predictor.compute_activity(5, pairs_45)[0], 0)
        self.assertGreater(self.predictor.compute_activity(5, pairs_45)[1], 0)
        self.assertGreater(self.predictor.compute_activity(5, pairs_54)[0], 0)
        self.assertEqual(self.predictor.compute_activity(5, pairs_54)[1], 0)
        self.assertListEqual(self.predictor.determine_highly_active(5, pairs_45),
                [False, True])
        self.assertListEqual(self.predictor.determine_highly_active(5, pairs_54),
                [True, False])


class TestSimpleBinaryPredictor(unittest.TestCase):
    """Tests methods in the SimpleBinaryPredictor class.
    """

    def test_compute_activity(self):
        seqs = ['ATCGAA',
                'ATCGAA',
                'GGGCCC',
                'ATCGAA',
                'AT-GAA',
                'ATCAAA',
                'GGGCCC']
        seqs_aln = alignment.Alignment.from_list_of_seqs(seqs)

        predictor = predict_activity.SimpleBinaryPredictor(1, False)
        activities = predictor.compute_activity(1, 'TCG', seqs_aln)
        expected = np.array([1, 1, 0, 1, 0, 1, 0])

        np.testing.assert_array_equal(activities, expected)




"""Utilities involving logging.
"""

import logging

__author__ = 'Hayden Metsky <hayden@mit.edu>'


def configure_logging(level=logging.WARNING):
    """Configure the global logging package.
    Args:
        level: logging level below which logging messages are ignored
    """
    format = '%(asctime)s - %(name)s [%(levelname)s] %(message)s'
    logging.basicConfig(format=format, level=level)



#!/usr/bin/python
"""This determines the adapt package version, primarily based on git.
If git is available, we use the version specified by git. This can indicate
commits on top of the last numbered version and can also indicate if the
working directory is dirty (i.e., has local modifications). If git is not
available but some version from git was stored in a file, we use that. Finally,
if none of these are available, we resort to the numbered release version
manually specified in the variable RELEASE_VERSION.
"""

# Manually specify a numbered release version to use as a fallback
RELEASE_VERSION = 'v1.2.0'


import subprocess
import os

__author__ = ['Danny Park <dpark@broadinstitute.org>',
        'Hayden Metsky <hayden@mit.edu>']

# Set __version__ lazily below
__version__ = None


def get_project_path():
    """Determine absolute path to the top-level of the project.
    This is assumed to be the parent of the directory containing this script.
    Returns:
        path (string) to top-level of the project
    """
    # abspath converts relative to absolute path; expanduser interprets ~
    path = __file__  # path to this script
    path = os.path.expanduser(path)  # interpret ~
    path = os.path.abspath(path)  # convert to absolute path
    path = os.path.dirname(path)  # containing directory: utils
    path = os.path.dirname(path)  # containing directory: project dir
    return path


def get_version_from_git_describe():
    """Determine a version according to git.
    This calls `git describe`, if git is available.
    Returns:
        version from `git describe --tags --always --dirty` if git is
        available; otherwise, None
    """
    cwd = os.getcwd()
    try:
        os.chdir(get_project_path())
        cmd = ['git', 'describe', '--tags', '--always', '--dirty']
        out = subprocess.check_output(cmd, stderr=subprocess.DEVNULL)
        if not isinstance(out, str):
            out = out.decode('utf-8')
        ver = out.strip()
    except Exception:
        ver = None
    os.chdir(cwd)
    return ver


def release_file():
    """Obtain path to file storing version, according to git.
    Returns:
        path to VERSION file
    """
    return os.path.join(get_project_path(), 'VERSION')


def read_release_version():
    """Read VERSION file, containing git version.
    Returns:
        if VERSION file exists, version stored in it; otherwise, None
    """
    try:
        with open(release_file(), 'rt') as inf:
            version = inf.readlines()[0].strip()
    except Exception:
        version = None
    return version


def write_release_version(version):
    """Save version, according to git, into VERSION file.
    Args:
        version: version to save
    """
    with open(release_file(), 'wt') as outf:
        outf.write(version + '\n')


def get_version():
    """Determine version from git, and save if available.
    """
    # Allow modifying the global __version__ variable
    global __version__

    if __version__ is None:
        from_git = get_version_from_git_describe()
        from_file = read_release_version()

        if from_git:
            # A version is available from git; use this
            if from_file != from_git:
                # Update the version stored in the VERSION file
                write_release_version(from_git)
            __version__ = from_git
        else:
            if from_file:
                # No version is available from git but one is in the
                # VERSION file; use this
                __version__ = from_file

        if __version__ is None:
            # No version is available from git and there is no VERSION
            # file; use the manually set release version
            __version__ = RELEASE_VERSION

    return __version__


def get_latest_model_version(model_path):
    """Get latest model version, given the model path
    """
    # List all model versions in path
    model_versions = os.listdir(model_path)
    # Get a list of the versions
    # Each version is represented as a list of numbers
    model_versions_numeric = []
    for model_version in model_versions:
        if model_version.startswith('v'):
            model_version_numeric = []
            skip = False
            for i in model_version[1:].split('_'):
                if not i.isdecimal():
                    skip = True
                    break
                else:
                    model_version_numeric.append(int(i))
            if not skip and len(model_version_numeric) > 0:
                model_versions_numeric.append(model_version_numeric)

    # If there were no models found on the path, raise an error
    if len(model_versions_numeric) == 0:
        raise ValueError("There are no appropriately formatted models in the "
            "model path. Please make sure the models are in a folder with the "
            "format 'v_#_#'")

    # Remake the version string
    latest_version = [str(i) for i in sorted(model_versions_numeric)[-1]]
    return 'v' + '_'.join(latest_version)


if __name__ == "__main__":
    # Determine and print the package version
    print(get_version())



"""Functions for performing an alignment of sequences, with curation.
"""

from collections import OrderedDict
import hashlib
import logging
import os
import random
import re
import subprocess
import tempfile
try:
    import boto3
    from botocore.exceptions import ClientError
except:
    cloud = False
else:
    cloud = True

from adapt.utils import seq_io

__author__ = 'Hayden Metsky <hayden@mit.edu>'

logger = logging.getLogger(__name__)


def read_unaligned_seqs(tmp):
    """Parse temp fasta file of sequences.

    Args:
        tmp: tempfile, which has a name attribute
            (tempfile.NamedTemporaryFile object)

    Returns:
        dict mapping accession.version to sequences (as str)
    """
    logger.debug(("Reading and parsing unaligned sequences from "
        "tmp file at %s") % tmp.name)

    # Each sequence header is in the format '[accession].[version] [name, etc.]'
    # Parse them to extract the accession.version (everything before the
    # first space)
    accver_p = re.compile('([^\s]+)')

    seqs = seq_io.read_fasta(tmp.name)
    seqs_by_accver = OrderedDict()
    for name, seq in seqs.items():
        accver = accver_p.match(name).group(1)
        seqs_by_accver[accver] = seq
    return seqs_by_accver


class AlignmentMemoizer:
    """Memoizer for alignments.

    This stores an alignment using a hash of the accession.version
    in the alignments.

    This stores alignments as fasta files named by the hash.
    """

    def __init__(self, path, aws_access_key_id=None, aws_secret_access_key=None):
        """
            Args:
                path: path to directory in which to read and store
                    memoized alignments; if path begins with "s3://",
                    stores bucket
        """
        if path[:5] == "s3://":
            # Store path as an S3 Bucket Object
            folders = path.split("/")
            self.path = "/".join(folders[3:])
            self.bucket = folders[2]
            if aws_access_key_id is not None and aws_secret_access_key is not None:
                self.S3 = boto3.client("s3", 
                    aws_access_key_id=aws_access_key_id, 
                    aws_secret_access_key=aws_secret_access_key)
            else:
                self.S3 = boto3.client("s3")
        else:
            self.path = path
            self.bucket = None
            if not os.path.exists(self.path):
                os.makedirs(self.path)

    def _hash_accvers_filepath(self, accvers):
        """Generate a path or S3 key to use as the filename for an alignment.

        Args:
            accvers: collection of accession.version in an alignment

        Returns:
            path to file or S3 key that should store accvers
        """
        h = hashlib.md5(str(sorted(set(accvers))).encode('utf-8')).hexdigest()
        if self.bucket:
            return "/".join([self.path, h])
        return os.path.join(self.path, h)

    def get(self, accvers):
        """Get memoized alignment.

        Args:
            accvers: collection of accession.version in an alignment

        Returns:
            OrderedDict mapping accession.version to sequences; or None
            if accvers is not memoized
        """
        p = self._hash_accvers_filepath(accvers)
        seqs = None

        if self.bucket:
            # Download file from source if it exists,
            # otherwise return None
            try:
                f = self.S3.get_object(Bucket = self.bucket, Key = p)
            except ClientError as e:
                if e.response['Error']['Code'] == "NoSuchKey":
                    return None
                else:
                    raise
            else:
                lines = [line.decode('utf-8') for line in f["Body"].iter_lines()]
                seqs = seq_io.process_fasta(lines)

        elif os.path.exists(p):
            # Read the fasta
            seqs = seq_io.read_fasta(p)

        if seqs:
            # Verify that the accessions in the file match accs 
            # (i.e., that there is not a collision)
            if set(accvers) == set(seqs.keys()):
                return seqs
            else:
                return None
        else:
            return None

    def save(self, seqs):
        """Memoize alignment by saving it to a file.

        Args:
            seqs: dict mapping accession.version to sequences
        """
        p = self._hash_accvers_filepath(seqs.keys())

        if self.bucket:
            # Make temporary file to store FASTA, then upload to S3
            p_tmp = tempfile.NamedTemporaryFile(delete=False)
            p_tmp.close()
            seq_io.write_fasta(seqs, p_tmp.name)
            with open(p_tmp.name, 'rb') as f:
                self.S3.put_object(Bucket=self.bucket, Key=p, Body=f)
            os.unlink(p_tmp.name)
        else:
            # Generate a random 8-digit hex to append to p temporarily, so that
            # two files don't write to p at the same time
            p_tmp = "%s.%08x" % (p, random.getrandbits(32))
            seq_io.write_fasta(seqs, p_tmp)
            os.replace(p_tmp, p)
        

class AlignmentStatMemoizer:
    """Memoizer for statistics on alignments.

    This stores an alignment using a hash of accession.version
    in the alignments.

    This stores, for each alignment, the tuple (aln_identity, aln_identity_ccg).
    """

    def __init__(self, path, aws_access_key_id=None, aws_secret_access_key=None):
        """
            Args:
                path: path to directory in which to read and store memoized
                    stats; if path begins with "s3://", stores bucket
        """
        if path[:5] == "s3://":
            # Store path as an S3 Bucket Object
            folders = path.split("/")
            self.path = "/".join(folders[3:])
            self.bucket = folders[2]
            if aws_access_key_id is not None and aws_secret_access_key is not None:
                self.S3 = boto3.client("s3", 
                    aws_access_key_id=aws_access_key_id, 
                    aws_secret_access_key=aws_secret_access_key)
            else:
                self.S3 = boto3.client("s3")
        else:
            self.path = path
            self.bucket = None
            if not os.path.exists(self.path):
                os.makedirs(self.path)

    def _hash_accvers_filepath(self, accvers):
        """Generate a path or S3 key to use as the filename for an alignment.

        Let h be the hash of accers and and h_2 be the first two
        (hexadecimal) digits of h. This stores h in a directory
        corresponding to h_2, in order to decrease the number of
        files per directory.

        Args:
            accvers: collection of accession.version in an alignment

        Returns:
            path to file or S3 key that should store stats on accvers
        """
        h = hashlib.md5(str(sorted(set(accvers))).encode('utf-8')).hexdigest()

        h_2 = h[:2]

        if self.bucket:
            return "/".join([self.path, h_2, h])

        h_dir = os.path.join(self.path, h_2)
        if not os.path.exists(h_dir):
            os.makedirs(h_dir)

        return os.path.join(h_dir, h)

    def get(self, accvers):
        """Get memoized stats for alignment.

        Args:
            accvers: collection of accession.version in an alignment

        Returns:
            (aln_identity, aln_identity_ccg) for alignment of
            accvers; or None if accvers is not memoized
        """
        p = self._hash_accvers_filepath(accvers)
        ls = None

        if self.bucket:
            # Download file from source if it exists,
            # otherwise return None
            try:
                f = self.S3.get_object(Bucket = self.bucket, Key = p)
            except ClientError as e:
                if e.response['Error']['Code'] == 'NoSuchKey':
                    return None
                else:
                    raise e
            else:
                lines = f["Body"].read().decode('utf-8').rstrip()
                ls = lines.split('\t')

        elif os.path.exists(p):
            # Read the file
            with open(p) as f:
                lines = [line.rstrip() for line in f]
            assert len(lines) == 1  # should be just 1 line
            ls = lines[0].split('\t')
        
        if ls:
            # Verify that the accessions in file match accvers
            # (i.e., that there is not a collision)
            accvers_in_p = ls[0].split(',')
            if set(accvers) == set(accvers_in_p):
                stats = (float(ls[1]), float(ls[2]))
                return stats
            else:
                return None
        else:
            return None

    def save(self, accvers, stats):
        """Memoize statistics for an alignment.

        Args:
            accvers: collection of accession.version in an alignment
            stats: tuple (aln_identity, aln_identity_ccg)
        """
        p = self._hash_accvers_filepath(accvers)
        accvers_str = ','.join(accvers)

        if self.bucket:
            body = '\t'.join([accvers_str, str(stats[0]), str(stats[1])]) + '\n'
            self.S3.put_object(Bucket=self.bucket, Key=p,
                Body=body.encode('utf-8'))
        else:
            # Generate a random 8-digit hex to append to p temporarily, so that
            # two files don't write to p at the same time
            p_tmp = "%s.%08x" % (p, random.getrandbits(32))
            with open(p_tmp, 'w') as fw:
                fw.write(('\t'.join([accvers_str, str(stats[0]), str(stats[1])]) +
                    '\n'))
            os.rename(p_tmp, p)


_mafft_exec = None
def set_mafft_exec(mafft_path):
    """Set path to the mafft executable.

    Args:
        mafft_path: path to mafft executable
    """
    global _mafft_exec
    # Verify the path exists and is an executable
    if not os.path.isfile(mafft_path) or not os.access(mafft_path, os.X_OK):
        raise FileNotFoundError(("Path to mafft (%s) does not exist or is not "
            "executable") % mafft_path)
    _mafft_exec = mafft_path


def align(seqs, am=None):
    """Align sequences using mafft.

    Args:
        seqs: dict mapping sequence header to sequences
        am: AlignmentMemoizer object to use for memoizing alignments;
            or None to not memoize

    Returns:
        dict mapping sequence header to sequences, where all the sequences
        are aligned
    """
    if am is not None:
        seqs_aligned = am.get(seqs.keys())
        if seqs_aligned is not None:
            return seqs_aligned

    if len(seqs) == 1:
        # There's one sequence; simply output it
        seqs_aligned = OrderedDict(seqs)
        if am is not None:
            am.save(seqs_aligned)
        return OrderedDict(seqs_aligned)

    global _mafft_exec
    # Check that mafft executable has been set
    if _mafft_exec is None:
        raise Exception(("Path to mafft executable must be set using "
            "_set_mafft_exec()"))

    # Write a fasta of these sequences to a named temp file, and
    # create one to write the output fasta to
    # Files closed for Windows, which will not allow files to be 
    # opened twice (files opened again in write_fasta/read_fasta)
    in_fasta = tempfile.NamedTemporaryFile(delete=False)
    in_fasta.close()
    seq_io.write_fasta(seqs, in_fasta.name)
    out_fasta = tempfile.NamedTemporaryFile(delete=False)

    # Setup arguments to mafft
    params = ['--preservecase', '--thread', '-1']
    max_seq_len = max(len(seq) for seq in seqs.values())
    if len(seqs) < 10 and max_seq_len < 10000:
        # Accuracy oriented
        params += ['--maxiterate', '1000', '--localpair']
    elif len(seqs) > 50000 or max_seq_len > 200000:
        # Speed oriented
        params += ['--retree', '1', '--maxiterate', '0']
    elif len(seqs) > 1000 or max_seq_len > 100000:
        # Speed oriented
        params += ['--retree', '2', '--maxiterate', '0']
    else:
        # Standard
        params += ['--retree', '2', '--maxiterate', '2']

    # Call mafft
    cmd = [_mafft_exec] + params + [in_fasta.name]
    subprocess.call(cmd, stdout=out_fasta, stderr=subprocess.DEVNULL)
    out_fasta.close()

    # Read the output fasta into a dict
    seqs_aligned = seq_io.read_fasta(out_fasta.name)

    # Delete temp files
    os.unlink(in_fasta.name)
    os.unlink(out_fasta.name)

    if len(seqs) > 0 and len(seqs_aligned) == 0:
        logger.critical(("The generated alignment contains no sequences; "
            "it is possible that mafft failed, e.g., due to running out "
            "of memory"))

    if am is not None:
        # Memoize the alignment
        am.save(seqs_aligned)

    return seqs_aligned


def _aln_identity(a, b):
    """Compute identity of alignment between two sequences.

    Args:
        a, b: two aligned sequences

    Returns:
        float representing the fraction of the alignment that is
        identical between a and b
    """
    assert len(a) == len(b)
    identical_count = sum(1 for i in range(len(a)) if a[i] == b[i])
    return identical_count / float(len(a))


def _collapse_consecutive_gaps(a, b):
    """Collapse consecutive gaps in an alignment between two sequences.

    For example, the alignment
        ATC----GA
        ATCATCGGA
    would become
        ATC-GA
        ATCAGA

    Args:
        a, b: two aligned sequences

    Returns:
        tuple (a', b') where a' and b' represents an alignment of a
        and b, with gaps collapsed
    """
    assert len(a) == len(b)
    a_ccg, b_ccg = a[0], b[0]
    for i in range(1, len(a)):
        if a[i-1] == '-':
            if a[i] == '-':
                # Skip this position; we are already in a gap
                continue
            else:
                # We are out of the gap for a; be sure to include
                # this position
                a_ccg += a[i]
                b_ccg += b[i]
        elif b[i-1] == '-':
            if b[i] == '-':
                # Skip this position; we are already in a gap
                continue
            else:
                # We are out of the gap for b; be sure to include
                # this position
                a_ccg += a[i]
                b_ccg += b[i]
        else:
            a_ccg += a[i]
            b_ccg += b[i]
    return (a_ccg, b_ccg)


def curate_against_ref(seqs, ref_accs, asm=None,
        aln_identity_thres=0.5, aln_identity_ccg_thres=0.6,
        aln_identity_long_thres=(100000, 0.9, 0.9),
        remove_ref_accs=[]):
    """Curate sequences by aligning pairwise with reference sequences.

    This can make use of an object of AlignmentStatMemoizer to
    memoize values, since the pairwise alignment can be slow.

    This compares each sequence against a list of reference accessions,
    and keeps any that satisfies the criteria against any reference
    sequence. In other words, it filters out any sequences that fail to
    meet the criteria against all reference sequences.

    Args:
        seqs: dict mapping sequence accession to sequences
        ref_accs: list of accessions of reference sequences to use for curation;
            for each ref_acc in ref_accs, either itself or ref_acc.[version]
            for some version must be a key in seqs
        asm: AlignmentStatMemoizer to use for memoization of values; or
            None to not memoize
        aln_identity_thres: filter out any sequences with less than
            this fraction identity to all ref_acc in ref_accs (taken over the
            whole alignment)
        aln_identity_ccg_thres: filter out any sequences with less than
            this fraction identity to all ref_acc in ref_accs, after collapsing
            consecutive gaps to a single gap; this should be at least
            aln_identity_thres
        aln_identity_long_thres: if set, then adjust aln_identity_thres
            and aln_identity_ccf_thres for long genomes. This is a tuple
            (len, a, b) such that, if the length of the longest reference
            sequence is >= len, then aln_identity_thres takes on the
            value a and aln_identity_ccf_thres takes on the value b. This
            is useful to be stronger in curating long genomes (ideally,
            removing sequences with large structural changes), which will
            take long to align in a multiple sequence alignment
        remove_ref_accs: a list of ref_acc, specifying to not include each
            ref_acc in the output

    Returns:
        dict mapping sequence accession.version to sequences, filtered by
        ones that can align reasonably well with some ref_acc in ref_accs
    """
    logger.info(("Curating %d sequences against references %s "
        "(this can take a while if alignment stats are not memoized)") %
        (len(seqs), ref_accs))

    # Find each ref_acc in seqs
    ref_acc_to_key = {}
    max_ref_seq_len = 0
    for ref_acc in ref_accs:
        if ref_acc in seqs:
            ref_acc_key = ref_acc
        else:
            # Look for a key in seqs that is ref_acc.[version] for some version
            ref_acc_key = None
            for k in seqs.keys():
                if k.split('.')[0] == ref_acc:
                    ref_acc_key = k
                    break
            if ref_acc_key is None:
                raise Exception(("Each ref_acc must be in seqs; it is possible "
                    "that the reference accession '%s' is invalid or could not "
                    "be found") % ref_acc)
        ref_acc_to_key[ref_acc] = ref_acc_key
        max_ref_seq_len = max(max_ref_seq_len, len(seqs[ref_acc_key]))

    if aln_identity_long_thres is not None:
        len_thres, a, b = aln_identity_long_thres
        if max_ref_seq_len >= len_thres:
            # This is a long sequence; be more strict in curation
            aln_identity_thres = a
            aln_identity_ccf_thres = b

    seqs_filtered = OrderedDict()
    for accver, seq in seqs.items():
        for ref_acc in ref_accs:
            ref_acc_key = ref_acc_to_key[ref_acc]
            if accver == ref_acc_key:
                # ref_acc_key will align well with ref_acc_key, so include it
                # automatically
                seqs_filtered[accver] = seq
                break

            stats = asm.get((ref_acc_key, accver)) if asm is not None else None
            if stats is not None:
                aln_identity, aln_identity_ccg = stats
            else:
                # Align ref_acc_key with accver
                to_align = {ref_acc_key: seqs[ref_acc_key], accver: seq}
                aligned = align(to_align)
                ref_acc_aln = aligned[ref_acc_key]
                accver_aln = aligned[accver]
                assert len(ref_acc_aln) == len(accver_aln)

                # Compute the identity of the alignment
                aln_identity = _aln_identity(ref_acc_aln, accver_aln)

                # Compute the identity of the alignment, after collapsing
                # consecutive gaps (ccg) to a single gap
                ref_acc_aln_ccg, accver_aln_ccg = _collapse_consecutive_gaps(
                    ref_acc_aln, accver_aln)
                aln_identity_ccg = _aln_identity(ref_acc_aln_ccg, accver_aln_ccg)

                if asm is not None:
                    # Memoize these statistics, so the alignment does not need
                    # to be performed again
                    stats = (aln_identity, aln_identity_ccg)
                    asm.save((ref_acc_key, accver), stats)

            logger.debug(("Alignment between %s and reference %s has identity "
                "%f and identity (after collapsing consecutive gaps) %f") %
                (accver, ref_acc_key, aln_identity, aln_identity_ccg))

            if (aln_identity >= aln_identity_thres and
                    aln_identity_ccg >= aln_identity_ccg_thres):
                # Include accver in the filtered output
                seqs_filtered[accver] = seq
                break

    logger.info(("After curation, %d of %d sequences (with unique accession) "
        "were kept; %d of these are references that will be removed"),
        len(seqs_filtered), len(seqs), len(remove_ref_accs))

    for remove_ref_acc in remove_ref_accs:
        # Do not include the ref_acc_key corresponding to remove_ref_acc
        # in the output
        ref_acc_key = ref_acc_to_key[remove_ref_acc]
        del seqs_filtered[ref_acc_key]

    return seqs_filtered



"""Tests for cluster module.
"""

import random
import unittest

import numpy as np
import scipy

from adapt.prepare import cluster

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestClusterFromMatrix(unittest.TestCase):
    """Test basic clustering functions."""

    def test_create_condensed_dist_matrix(self):
        # Have 3 elements: 0 and 1 are similar, 0 and 2 are
        # very dissimilar, and 1 and 2 are similar
        n = 3
        dist_matrix_2d = np.array(
            [[0,   1, 100],
             [1,   0, 2  ],
             [100, 2, 0  ]])
        def dist_fn(i, j):
            return dist_matrix_2d[i][j]

        condensed = cluster.create_condensed_dist_matrix(n, dist_fn)

        # Use scipy to create a condensed matrix, which is possible
        # since we already have the 2d matrix available
        scipy_condensed = scipy.spatial.distance.squareform(dist_matrix_2d)

        self.assertTrue(np.array_equal(condensed, scipy_condensed))

    def test_cluster_from_dist_matrix_1(self):
        # Have 3 elements: 0 and 1 are similar, 0 and 2 are
        # very dissimilar, and 1 and 2 are similar
        n = 3
        dists = {(0, 1): 1, (0, 2): 100, (1, 2): 2}
        def dist_fn(i, j):
            return dists[(i, j)]
        dist_matrix = cluster.create_condensed_dist_matrix(n, dist_fn)

        clusters = cluster.cluster_from_dist_matrix(dist_matrix, 10)
        self.assertEqual(clusters, [[0, 1], [2]])

    def test_cluster_from_dist_matrix_2(self):
        # Have 3 elements: 0 and 1 are very dissimilar similar, 0 and 2 are
        # very dissimilar, and 1 and 2 are similar
        n = 3
        dists = {(0, 1): 100, (0, 2): 100, (1, 2): 1}
        def dist_fn(i, j):
            return dists[(i, j)]
        dist_matrix = cluster.create_condensed_dist_matrix(n, dist_fn)

        clusters = cluster.cluster_from_dist_matrix(dist_matrix, 10)
        self.assertEqual(clusters, [[1, 2], [0]])

    def test_cluster_from_dist_matrix_3(self):
        # Have 3 elements: 0 and 1 are very dissimilar similar, 0 and 2 are
        # very dissimilar, and 1 and 2 are similar
        # Have 3 elements: all very dissimilar
        n = 3
        dists = {(0, 1): 20, (0, 2): 30, (1, 2): 40}
        def dist_fn(i, j):
            return dists[(i, j)]
        dist_matrix = cluster.create_condensed_dist_matrix(n, dist_fn)

        clusters = cluster.cluster_from_dist_matrix(dist_matrix, 10)
        self.assertEqual(sorted(clusters), [[0], [1], [2]])


class TestClusterWithMinHashSignatures(unittest.TestCase):
    """Test cluster_with_minhash_signatures() function."""

    def test_simple(self):
        seqs = {'a': 'AT'*500,
                'b': 'CG'*500,
                'c': 'AT'*500,
                'd': 'TA'*500,
                'e': 'TT'*500,
                'f': 'CG'*500}
        # The expected clusters should be: [a, c, d], [b, f], [e]

        clusters = cluster.cluster_with_minhash_signatures(seqs)
        self.assertEqual(len(clusters), 3)
        self.assertEqual(sorted(clusters[0]), ['a', 'c', 'd'])
        self.assertEqual(sorted(clusters[1]), ['b', 'f'])
        self.assertEqual(sorted(clusters[2]), ['e'])

class TestFindRepresentativeSequences(unittest.TestCase):
    """Test find_representative_sequences() function."""

    def test_simple(self):
        random.seed(1)
        np.random.seed(1)

        seqs = {'a': 'ATCGAATTCGGATCG',
                'b': 'CCCCCCCCCCCCCCC',
                'c': 'ATCGTATTCGGATCG',
                'd': 'ATCGTATTCGGTTCG'}
        # The expected clusters should be [a, c, d], [b]
        # The medoid of the [a,c,d] cluster should be c

        rep_seqs, _ = cluster.find_representative_sequences(
                seqs, k=3, N=12, threshold=0.3)
        self.assertCountEqual(rep_seqs, {'c', 'b'})

    def test_num_clusters(self):
        random.seed(1)
        np.random.seed(1)

        seqs = {'a': 'ATCGAATTCGGATCG',
                'b': 'CCCCCCCCCCCCCCC',
                'c': 'ATCGTATTCGGATCG',
                'd': 'ATCGTATTCGGTTCG'}

        # The expected clusters should be [a, c, d], [b]
        # The medoid of the [a,c,d] cluster should be c
        rep_seqs, _ = cluster.find_representative_sequences(
                seqs, k=3, N=12, threshold=None, num_clusters=2)
        self.assertCountEqual(rep_seqs, {'c', 'b'})

        # The expected cluster should be [a, b, c, d] with medoid c
        rep_seqs, _ = cluster.find_representative_sequences(
                seqs, k=3, N=12, threshold=None, num_clusters=1)
        self.assertCountEqual(rep_seqs, {'c'})



"""Tests for coverage_analysis module.
"""

import logging
import unittest

from adapt import coverage_analysis

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestCoverageAnalysisWithMismatchModel(unittest.TestCase):
    """Tests methods in the CoverageAnalysisWithMismatchModel class.
    """

    def setUp(self):
        # Disable logging
        logging.disable(logging.WARNING)

        self.seqs = {
                'seq1': 'ATCGAATTCGATCGAA',
                'seq2': 'GGATCGAATTCGAG',
                'seq3': 'TGCAACCGATCCGT'
        }
        self.design1 = coverage_analysis.Design({'TTCGA'})
        self.design2 = coverage_analysis.Design({'TTCGA'},
                ({'TCGA'}, {'CGAT'}))
        self.design3 = coverage_analysis.Design({'TTCGA', 'CCGA', 'GGGGG'},
                ({'TCGA', 'GCAA', 'AAAA'}, {'CGAT', 'TCCG', 'CCCC'}))
        designs = {
                'design1': self.design1,
                'design2': self.design2,
                'design3': self.design3
        }
        self.ca = coverage_analysis.CoverageAnalyzerWithMismatchModel(self.seqs,
                designs, 1, 1, allow_gu_pairs=False)

        # Index sequences with a k-mer length of k=2 to ensure k-mers
        # will be found
        self.ca._index_seqs(k=2, stride_by_k=False)

    def test_find_binding_pos(self):
        bind_fn = self.ca.guide_bind_fn
        for fully_sensitive in [False, True]:
            self.ca.fully_sensitive = fully_sensitive
            self.assertEqual(
                    self.ca.find_binding_pos('seq1', 'TTCGAT', bind_fn),
                    {6}
            )
            self.assertEqual(
                    self.ca.find_binding_pos('seq1', 'TTCC', bind_fn),
                    {6}
            )
            self.assertEqual(
                    self.ca.find_binding_pos('seq1', 'AAAA', bind_fn),
                    set()
            )
            if fully_sensitive is True:
                self.assertEqual(
                        self.ca.find_binding_pos('seq1', 'AT', bind_fn),
                        {0,4,5,6,10,14}
                )
        self.ca.fully_sensitive = False

    def test_seqs_where_guide_binds(self):
        self.assertEqual(
                self.ca.seqs_where_guides_bind({'TTCGA'}),
                {'seq1', 'seq2'}
        )
        self.assertEqual(
                self.ca.seqs_where_guides_bind({'TTCGA', 'CCGA'}),
                {'seq1', 'seq2', 'seq3'}
        )
        self.assertEqual(
                self.ca.seqs_where_guides_bind({'GGGCC'}),
                set()
        )

    def test_seqs_where_targets_bind(self):
        self.assertEqual(
                self.ca.seqs_where_targets_bind({'TTCGA'},
                    {'TCGA'}, {'CGAT'}),
                {'seq1'}
        )

    def test_seqs_bound_by_design(self):
        self.assertEqual(self.ca.seqs_bound_by_design(self.design1),
                {'seq1', 'seq2'})
        self.assertEqual(self.ca.seqs_bound_by_design(self.design2),
                {'seq1'})
        self.assertEqual(self.ca.seqs_bound_by_design(self.design3),
                {'seq1', 'seq3'})
        
    def test_frac_of_seqs_bound(self):
        self.assertEqual(self.ca.frac_of_seqs_bound(),
                {'design1': 2.0/3.0, 'design2': 1.0/3.0, 'design3': 2.0/3.0})

    def tearDown(self):
        # Re-enable logging
        logging.disable(logging.NOTSET)


class TestCoverageAnalysisWithPredictedActivity(unittest.TestCase):
    """Tests methods in the CoverageAnalysisWithPredictedActivity class.
    """

    def setUp(self):
        # Disable logging
        logging.disable(logging.WARNING)

        # Predict guides matching target to have activity 1, and
        # starting with 'A' to have activity 2 (otherwise, 0)
        # Predict guides to be highly active iff they start with A and
        # and match target
        class PredictorTest:
            def __init__(self):
                self.context_nt = 1
            def determine_highly_active(self, start_pos, pairs):
                y = []
                for target, guide in pairs:
                    target_without_context = target[self.context_nt:len(target)-self.context_nt]
                    if guide == target_without_context:
                        if guide[0] == 'A':
                            y += [True]
                        else:
                            y += [False]
                    else:
                        y += [False]
                return y
            def compute_activity(self, start_pos, pairs):
                y = []
                for target, guide in pairs:
                    target_without_context = target[self.context_nt:len(target)-self.context_nt]
                    if guide == target_without_context:
                        if guide[0] == 'A':
                            y += [2]
                        else:
                            y += [1]
                    else:
                        y += [0]
                return y
            def cleanup_memoized(self, pos):
                pass
        predictor = PredictorTest()

        self.seqs = {
                'seq1': 'ATCGAATTCGATCGAA',
                'seq2': 'GGATCGAATTCGAG',
                'seq3': 'TGCAACCGATCCGT'
        }
        self.ca = coverage_analysis.CoverageAnalyzerWithPredictedActivity(
                self.seqs, {}, predictor, 1, highly_active=True)

        # Index sequences with a k-mer length of k=2 to ensure k-mers
        # will be found
        self.ca._index_seqs(k=2, stride_by_k=False)

    def test_seqs_where_guide_binds(self):
        self.assertEqual(
                self.ca.seqs_where_guides_bind({'TTCGA'}),
                set()
        )
        self.assertEqual(
                self.ca.seqs_where_guides_bind({'TTCGA', 'CCGA'}),
                set()
        )
        self.assertEqual(
                self.ca.seqs_where_guides_bind({'AATT'}),
                {'seq1', 'seq2'}
        )
        self.assertEqual(
                self.ca.seqs_where_guides_bind({'GGGCC'}),
                set()
        )




#!/usr/bin/env python3
"""Pick targets recommended for testing."""

import argparse
import logging

from adapt import alignment
from adapt.prepare import cluster
from adapt import target_search
from adapt.utils import formatting
from adapt.utils import guide
from adapt.utils import log
from adapt.utils import seq_io

__author__ = 'Hayden Metsky <hayden@mit.edu>'

logger = logging.getLogger(__name__)


def find_test_targets(design_target, aln, args):
    """Find collection of test target sequences for a design option.

    Note that there is probably going to be some confusion with the word
    'target' here. It refers both to a design option (in the TSV written
    by design.py), as well as a sequence that should be used for testing.

    Args:
        design_target: target_search.DesignTarget object, representing
            a design option (amplicon)
        aln: alignment of genomes; positions correspond to positions in
            design_target
        args: parsed arguments

    Returns:
        collection of tuples (s, s_tagged, f) where s is a target sequence
        to use for testing; s_tagged is s tagged to show where
        primers and guides may bind s; and f is the fraction of all sequences
        that s represents
    """
    logger.info(("Finding test targets for design option with endpoints "
        "[%d, %d)"), design_target.target_start, design_target.target_end)
    
    # Expand the extracted range to the minimum
    # Note that this extracts to get a target length of args.min_target_len
    # *in the alignment*; if there are gaps in the alignment within the
    # region, the actual length of the sequences could be shorter
    target_len = design_target.target_end - design_target.target_start
    nt_to_add = int(max(0, args.min_target_len - target_len) / 2)
    target_start = max(0, design_target.target_start - nt_to_add)
    target_end = min(aln.seq_length, design_target.target_end + nt_to_add)
    if target_end - target_start == args.min_target_len - 1:
        # Fix off-by-1 edge case
        target_end = min(aln.seq_length, target_end + 1)

    # Extract the alignment where this design target (amplicon) is
    aln_extract = aln.extract_range(target_start, target_end)
    # Pull out the sequences, without gaps
    aln_extract_seqs = aln_extract.make_list_of_seqs(remove_gaps=True)

    # Remove target sequences that are too short
    # This can happen due to gaps in the alignment; a sequence can have
    # length 0, for example, if is it all '-' in the amplicon (extracted
    # range)
    # If they are extremely short (shorter than minhash_k, below), then
    # this will cause an error downstream
    aln_extract_seqs = [s for s in aln_extract_seqs
            if len(s) >= args.min_seq_len_to_consider]

    # Add indices for each sequence so it can be used as a dict
    aln_extract_seqs_dict = {i: s for i, s in enumerate(aln_extract_seqs)}

    # The number of k-mers to use with MinHash for clustering cannot be
    # more than the number of k-mers in a sequence
    minhash_k = 12
    min_seq_len = min(len(s) for s in aln_extract_seqs)
    minhash_N = min(50, min_seq_len - minhash_k - 1)

    # Find representative sequences
    if args.num_representative_targets:
        # Use a maximum number of clusters, ignoring the inter-cluster
        # distance threshold
        threshold = None
        num_clusters = args.num_representative_targets

        if args.min_frac_to_cover_with_rep_seqs < 1.0:
            logger.warning(("Fewer than %d targets may be reported "
                "because --min-frac-to-cover-with-rep-seqs is <1.0; set "
                "it to 1.0 to obtain %d representative targets") %
                (num_clusters, num_clusters))
    else:
        # Use an inter-cluster distance threhsold
        threshold = args.max_cluster_distance
        num_clusters = None
    rep_seqs_idx, rep_seqs_frac = cluster.find_representative_sequences(
            aln_extract_seqs_dict,
            k=minhash_k, N=minhash_N, threshold=threshold,
            num_clusters=num_clusters,
            frac_to_cover=args.min_frac_to_cover_with_rep_seqs)
    rep_seqs = [aln_extract_seqs[i] for i in rep_seqs_idx]

    # Find where primers and guides overlap each representative sequence
    guide_allow_gu_pairs = True
    if args.do_not_allow_gu_pairing:
        guide_allow_gu_pairs = False
    rep_seqs_tagged = []
    for rep_seq in rep_seqs:
        primer_seqs = (design_target.left_primer_seqs +
                design_target.right_primer_seqs)
        primer_overlap = guide.guide_overlap_in_seq(primer_seqs,
                rep_seq, args.pm, False)
        guide_seqs = design_target.guide_seqs
        guide_overlap = guide.guide_overlap_in_seq(guide_seqs,
                rep_seq, args.gm, guide_allow_gu_pairs)
        overlap_labels = {'primer': primer_overlap, 'guide': guide_overlap}

        if len(primer_overlap) == 0 and len(guide_overlap) == 0:
            logger.warning(("Neither primers nor guides could be found "
                "in a representative target sequence; it is possible that "
                "the representative is from the wrong region, or it is "
                "too divergent from the primer/guide sequences (for the "
                "latter, try raising -pm and -gm). The representative target "
                "sequence is for a design from the window [%d, %d) of the input "
                "alignment. This only matters for tagging the target, and it "
                "is ok to ignore this message.") % (design_target.target_start,
                    design_target.target_end))
        elif len(primer_overlap) == 0:
            logger.warning(("Primers could not be found in a representative "
                "target sequence; it is possible that the representative "
                "is from the wrong region, or it is too divergent from the "
                "primer sequences (for the latter, try raising -pm). The "
                "representative target sequence is for a design from the "
                "window [%d, %d) of the input alignment. This only matters "
                "for tagging the target, and it is ok to ignore this message.") %
                (design_target.target_start, design_target.target_end))
        elif len(guide_overlap) == 0:
            logger.warning(("Guides could not be found in a representative "
                "target sequence; it is possible that the representative "
                "is from the wrong region, or it is too divergent from the "
                "guide sequences (for the latter, try raising -gm). The "
                "representative target sequence is for a design from the "
                "window [%d, %d) of the input alignment. This only matters "
                "for tagging the target, and it is ok to ignore this message.") %
                (design_target.target_start, design_target.target_end))

        rep_seq_tagged = formatting.tag_seq_overlap(overlap_labels, rep_seq)
        rep_seqs_tagged += [rep_seq_tagged]

    return list(zip(rep_seqs, rep_seqs_tagged, rep_seqs_frac))


def main(args):
    # Read the design targets
    targets = target_search.DesignTarget.read_design_targets(args.design_tsv)

    # Read the alignment
    seqs = seq_io.read_fasta(args.alignment_fasta)
    aln = alignment.Alignment.from_list_of_seqs(list(seqs.values()))

    # Find and write test targets
    with open(args.out_tsv, 'w') as f:
        def write_row(row):
            line = '\t'.join(str(x) for x in row)
            f.write(line + '\n')
        header = ['design-target-start', 'design-target-end',
                'test-target-frac-represented',
                'test-target-seq', 'test-target-seq-tagged']
        write_row(header)
        for design_target in targets:
            targets_to_test = find_test_targets(design_target, aln, args)
            for tt in targets_to_test:
                row = [design_target.target_start, design_target.target_end,
                        tt[2], tt[0], tt[1]]
                write_row(row)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument('design_tsv',
            help=("Path to TSV with design options, as output by design.py. "
                  "Alternatively, it can be a custom-made TSV, with one "
                  "row per design option, containing the following columns: "
                  "'target-start' (5' end of a targeted genomic region, e.g., "
                  "an amplicon); 'target-end' (3' end); 'guide-target-"
                  "sequences' (space-separated list of guide sequences); "
                  "'left-primer-target-sequences' (space-separated list of "
                  "forward primer sequences); 'right-primer-target-"
                  "sequences' (space-separated list of reverse primer "
                  "sequences)"))
    parser.add_argument('alignment_fasta',
        help=("Path to alignment (FASTA) from which to extract targets. "
              "Target positions in DESIGN_TSV must correspond to positions "
              "in this alignment. If DESIGN_TSV is the output of design.py, "
              "then this can be the output of `design.py --write-input-aln`."))
    parser.add_argument('out_tsv',
        help=("Path to output TSV with recommended targets for testing"))

    parser.add_argument('-pm',
        type=int, default=5,
        help=("Number of mismatches to tolerate when determining whether "
              "primer binds to a region of target sequence"))
    parser.add_argument('-gm',
        type=int, default=3,
        help=("Number of mismatches to tolerate when determining whether "
              "guide binds to a region of target sequence"))
    parser.add_argument('--do-not-allow-gu-pairing', action='store_true',
        help=("When determining whether a guide binds to a region of target "
              "sequence, do not count G-U (wobble) base pairs as matching."))

    parser.add_argument('--min-seq-len-to-consider',
        type=int, default=80,
        help=("Do not consider, when identifying representative sequences, "
              "target sequences that are shorter than this length. These "
              "can occur due to gaps in the alignment (e.g., a target "
              "sequence can have length 0 if it is all '-' in the amplicon."))
    parser.add_argument('--min-target-len',
        type=int, default=500,
        help=("Minimum length of a target region; if the region "
              "in a design bound by primers is less than this, sequence "
              "will be added on both sides of the primer to reach this "
              "length. Note that this is in the alignment; the actual "
              "sequence could be shorter if there are gaps in the "
              "alignment"))
    parser.add_argument('--min-frac-to-cover-with-rep-seqs',
        type=float, default=0.95,
        help=("For representative sequences, use medoids of clusters such "
              "that the clusters account for at least this fraction of all "
              "sequences. This allows ignoring outlier clusters (whose "
              "sequence(s) may have not been covered by the design."))

    parser.add_argument('--max-cluster-distance',
        type=float, default=0.1,
        help=("Maximum inter-cluster distance to merge clusters (measured "
              "as 1-ANI). Higher values result in fewer representative "
              "targets."))
    parser.add_argument('--num-representative-targets',
        type=int,
        help=("Maximum number of clusters (equivalent to maximum number "
              "of representative targets). If set, then "
              "--max-cluster-distance is ignored. Note that fewer may "
              "be reported if --min-frac-to-cover-with-rep-seqs is "
              "<1.0; set it to 1.0 to report all."))

    # Log levels
    parser.add_argument("--debug",
        dest="log_level",
        action="store_const",
        const=logging.DEBUG,
        default=logging.WARNING,
        help=("Debug output"))
    parser.add_argument("--verbose",
        dest="log_level",
        action="store_const",
        const=logging.INFO,
        help=("Verbose output"))

    args = parser.parse_args()
    log.configure_logging(args.log_level)

    main(args)



"""Tests mutate module.
"""

import unittest

import numpy as np

from adapt import alignment
from adapt.utils import mutate, predict_activity

__author__ = 'Priya Pillai <ppillai@broadinstitute.org>'


class TestGTRSubstitutionMutator(unittest.TestCase):
    """Tests methods in the GTRSubstitutionMutator class.
    """
    def setUp(self):
        self.seqs = ['AC',
                     'GT']
        aln_a = alignment.Alignment.from_list_of_seqs(self.seqs)
        self.mutator_a = mutate.GTRSubstitutionMutator(aln_a,
                                                       1, 1, 1, 1, 1, 1,
                                                       1, 1000, 10)
        self.wild_seq = 'AAAAAAAA'
        self.mutated_seqs = self.mutator_a.mutate(self.wild_seq)

        aln_b = alignment.Alignment.from_list_of_seqs(self.mutated_seqs)
        rates = np.random.rand(6)
        self.mutator_b = mutate.GTRSubstitutionMutator(aln_b,
                                                       *rates,
                                                       1, 1, 5)

    def test_construct_rateQ(self):
        expected_a_Q = [[-1, 1/3, 1/3, 1/3],
                        [1/3, -1, 1/3, 1/3],
                        [1/3, 1/3, -1, 1/3],
                        [1/3, 1/3, 1/3, -1]]
        np.testing.assert_allclose(self.mutator_a.Q, expected_a_Q)

        Q = self.mutator_b.Q
        for row in Q:
            self.assertAlmostEqual(sum(row), 0)

    def test_construct_P(self):
        expected_a_P = [[0.25, 0.25, 0.25, 0.25],
                        [0.25, 0.25, 0.25, 0.25],
                        [0.25, 0.25, 0.25, 0.25],
                        [0.25, 0.25, 0.25, 0.25]]
        np.testing.assert_allclose(self.mutator_a.P, expected_a_P)

        P = self.mutator_b.P
        for row in P:
            self.assertAlmostEqual(sum(row), 1)

    def test_mutate(self):
        self.assertEqual(len(self.mutated_seqs), self.mutator_a.n)
        for mutated_seq in self.mutated_seqs:
            self.assertEqual(len(mutated_seq), len(self.wild_seq))

    def test_compute_sequence_probability(self):
        for mutated_seq in self.mutated_seqs:
            self.assertAlmostEqual(self.mutator_a.compute_sequence_probability(
                self.wild_seq, mutated_seq), (0.25**8))

    def test_compute_mutated_activity(self):
        predictor_a = predict_activity.SimpleBinaryPredictor(8, False)
        activity = self.mutator_a.compute_mutated_activity(predictor_a,
                                                           self.wild_seq,
                                                           self.wild_seq)
        self.assertEqual(activity, 1)

        predictor_a.mismatches = 0
        predictor_a.required_flanking_seqs = ('A', None)
        activity = self.mutator_a.compute_mutated_activity(predictor_a,
                                                           'A'+self.wild_seq,
                                                           self.wild_seq)
        self.assertEqual(activity, 0)



"""Functions for clustering sequences before input.

This includes computing a distance matrix using MinHash, and
clustering that matrix.
"""

from collections import defaultdict
from collections import OrderedDict
import logging
import operator

import numpy as np
from scipy.cluster import hierarchy

from adapt.utils import lsh

__author__ = 'Hayden Metsky <hayden@mit.edu>'

logger = logging.getLogger(__name__)


def make_signatures_with_minhash(family, seqs):
    """Construct a signature using MinHash for each sequence.

    Args:
        family: lsh.MinHashFamily object
        seqs: dict mapping sequence header to sequences

    Returns:
        dict mapping sequence header to signature
    """
    # Construct a single hash function; use the same for all sequences
    h = family.make_h()

    signatures = {}
    for name, seq in seqs.items():
        signatures[name] = h(seq)
    return signatures


def create_condensed_dist_matrix(n, dist_fn):
    """Construct a 1d condensed distance matrix for scipy.

    Args:
        n: number of elements whose pairwise distances to store in the
            matrix
        dist_fn: function such that dist_fn(i, j) gives the distance
            between i and j, for all i<j<n

    Returns:
        condensed 1d distance matrix for input to scipy functions
    """
    def idx(i, j):
        # Compute index in 1d vector for pair (i, j)
        return int((-1 * i*i)/2 + i*n - 3*i/2 + j - 1)

    dist_matrix_len = int(n*(n-1)/2)
    dist_matrix = np.zeros(dist_matrix_len)

    for j in range(n):
        for i in range(j):
            dist_matrix[idx(i, j)] = dist_fn(i, j)

    return dist_matrix


def cluster_from_dist_matrix(dist_matrix, threshold, num_clusters=None):
    """Use scipy to cluster a distance matrix.

    Args:
        dist_matrix: distance matrix, represented in scipy's 1d condensed form
        threshold: maximum inter-cluster distance to merge clusters (higher
            results in fewer clusters)
        num_clusters: if set, cluster such that this is the maximum number
            of clusters; threshold is ignored and must be None

    Returns:
        list c such that c[i] is a collection of all the observations
        (whose pairwise distances are indexed in dist) in the i'th
        cluster, in sorted order by descending cluster size
    """
    if threshold is None and num_clusters is None:
        raise ValueError(("threshold or num_clusters must be set"))
    elif threshold is not None and num_clusters is not None:
        raise ValueError(("Only one of threshold or num_clusters can be "
            "set"))

    linkage = hierarchy.linkage(dist_matrix, method='average')
    if threshold:
        clusters = hierarchy.fcluster(linkage, threshold, criterion='distance')
    elif num_clusters:
        clusters = hierarchy.fcluster(linkage, num_clusters,
            criterion='maxclust')

    # clusters are numbered starting at 1, but base the count on
    # first_clust_num just in case this changes
    first_clust_num = min(clusters)
    num_clusters_found = max(clusters) + 1 - first_clust_num
    elements_in_cluster = defaultdict(list)
    for i, clust_num in enumerate(clusters):
        elements_in_cluster[clust_num].append(i)
    cluster_sizes = {c: len(elements_in_cluster[c])
                     for c in range(first_clust_num,
                                    num_clusters_found + first_clust_num)}

    elements_in_cluster_sorted = []
    for clust_num, _ in sorted(cluster_sizes.items(),
            key=operator.itemgetter(1), reverse=True):
        elements_in_cluster_sorted += [elements_in_cluster[clust_num]]
    return elements_in_cluster_sorted


def cluster_with_minhash_signatures(seqs, k=12, N=100, threshold=0.1,
        num_clusters=None, return_dist_matrix_and_indices=False):
    """Cluster sequences based on their MinHash signatures.

    Args:
        seqs: dict mapping sequence header to sequences
        k: k-mer size to use for k-mer hashes (smaller is likely more
            sensitive for divergent genomes, but may lead to false positives
            in determining which genomes are close)
        N: number of hash values to use in a signature (higher is slower for
            clustering, but likely more sensitive for divergent genomes)
        threshold: maximum inter-cluster distance to merge clusters, in
            average nucleotide dissimilarity (1-ANI, where ANI is
            average nucleotide identity); higher results in fewer
            clusters
        num_clusters: if set, cluster such that this is the maximum number
            of clusters; threshold is ignored and must be None
        return_dist_matrix_and_indices: if set, return the pairwise distance
            matrix and sequences indices in each cluster; when used,
            seqs should be an OrderedDict

    Returns:
        if return_dist_matrix_and_indices:
            tuple (dm, c) such that dm is the pairwise distance matrix and
            c is a list such that c[i] gives a collection of sequence
            indices (corresponding to indices in dm) in the same cluster
            (note dm is a 1d condensed matrix in scipy's form); clusters in
            c are sorted in descending order of size
        else:
            list c such that c[i] gives a collection of sequence headers
            in the same cluster, and the clusters in c are sorted
            in descending order of size
    """
    if len(seqs) == 1:
        # Simply return one cluster
        if return_dist_matrix_and_indices:
            return None, [[0]]
        else:
            return [[list(seqs.keys())[0]]]

    family = lsh.MinHashFamily(k, N=N)
    signatures_map = make_signatures_with_minhash(family, seqs)

    # Map each sequence header to an index (0-based), and get
    # the signature for the corresponding index
    num_seqs = len(seqs)
    seq_headers = []
    signatures = []
    for name, seq in seqs.items():
        seq_headers += [name]
        signatures += [signatures_map[name]]

    # Eq. 4 of the Mash paper (Ondov et al. 2016) shows that the
    # Mash distance, which is shown to be closely related to 1-ANI, is:
    #  D = (-1/k) * ln(2*j/(1+j))
    # where j is a Jaccard similarity. Solving for j:
    #  j = 1/(2*exp(k*D) - 1)
    # So, for a desired distance D in terms of 1-ANI, the corresponding
    # Jaccard distance is:
    #  1.0 - 1/(2*exp(k*D) - 1)
    # We can use this to calculate a clustering threshold in terms of
    # Jaccard distance
    if threshold is not None:
        jaccard_dist_threshold = 1.0 - 1.0/(2.0*np.exp(k*threshold) - 1)
    else:
        # Ignore inter-cluster distance; use num_clusters instead
        jaccard_dist_threshold = None

    def jaccard_dist(i, j):
        # Return estimated Jaccard dist between signatures at
        # index i and index j
        return family.estimate_jaccard_dist(
            signatures[i], signatures[j])

    dist_matrix = create_condensed_dist_matrix(num_seqs, jaccard_dist)
    clusters = cluster_from_dist_matrix(dist_matrix,
        jaccard_dist_threshold, num_clusters=num_clusters)

    if return_dist_matrix_and_indices:
        return dist_matrix, clusters
    else:
        seqs_in_cluster = []
        for cluster_idxs in clusters:
            seqs_in_cluster += [[seq_headers[i] for i in cluster_idxs]]
        return seqs_in_cluster


def find_representative_sequences(seqs, k=12, N=100, threshold=0.1,
        num_clusters=None, frac_to_cover=1.0):
    """Find a set of representative sequences.

    This clusters seqs, and then determines a medoid for each cluster.
    It returns the medoids.

    This will not return representative sequences with ambiguity or NNNs.

    Args:
        seqs, k, N, threshold, num_clusters: see cluster_with_minhash_signatures()
        frac_to_cover: return medoids from clusters that collectively
            account for at least this fraction of all sequences; this
            allows ignoring representative sequences for outlier
            clusters

    Returns:
        tuple (set of sequence headers representing cluster medoids,
        fraction of all sequences contained in cluster)
    """
    seqs = OrderedDict(seqs)
    dist_matrix, clusters = cluster_with_minhash_signatures(
            seqs, k=k, N=N, threshold=threshold, num_clusters=num_clusters,
            return_dist_matrix_and_indices=True)

    seqs_items = list(seqs.items())
    n = len(seqs)
    def idx(i, j):
        # Compute index in 1d vector for pair (i, j)
        if i > j:
            i, j = j, i
        return int((-1 * i*i)/2 + i*n - 3*i/2 + j - 1)

    rep_seqs = []
    rep_seqs_frac = []
    num_seqs_accounted_for = 0
    for cluster_idxs in clusters:
        # Stop if we have already accounted for frac_to_cover of the
        # sequences
        # Note that clusters should be sorted in descending order of
        # size, so any clusters after this one will be the same size
        # or smaller
        if float(num_seqs_accounted_for) / len(seqs) >= frac_to_cover:
            break

        # Find the medoid of this cluster
        # Simply look over all pairs in the cluster (there are faster
        # algorithms, though not linear)
        curr_medoid = None
        curr_medoid_dist_total = None
        for i in cluster_idxs:
            # Only allow i to be the medoid if it does not have ambiguity
            seq = seqs_items[i][1]
            if sum(seq.count(b) for b in ('A','C','G','T')) != len(seq):
                # Has ambiguity or NNNs; skip
                continue

            # Compute the total distance to all other sequences in this
            # cluster, and check if this is the medoid
            dist_total = 0
            for j in cluster_idxs:
                if i == j:
                    continue
                dist_total += dist_matrix[idx(i,j)]
            if curr_medoid is None or dist_total < curr_medoid_dist_total:
                curr_medoid = i
                curr_medoid_dist_total = dist_total
        if curr_medoid is not None:
            rep_seqs += [curr_medoid]
            rep_seqs_frac += [float(len(cluster_idxs)) / len(seqs)]
            num_seqs_accounted_for += len(cluster_idxs)
        else:
            # All sequences have ambiguity or NNNs; raise a warning and
            # skip this cluster
            logger.warning(("Cannot find medoid for cluster of size %d "
                "because all sequences have ambiguity or NNNs; skipping "
                "this cluster"),
                len(cluster_idxs))

    return ([seqs_items[i][0] for i in rep_seqs], rep_seqs_frac)



"""Setup script for the adapt distribution using Distutils.
"""

from setuptools import find_packages
from setuptools import setup

import adapt

__author__ = 'Hayden Metsky <hmetsky@broadinstitute.org>, Priya P. Pillai <ppillai@broadinstitute.org>'

with open("README.md", "r", encoding="utf-8") as f:
  LONG_DESCRIPTION = f.read()

# There is a PyPi package called 'adapt' already, so renamed for PyPi
# Does not affect the Bioconda package name
setup(name='adapt-diagnostics',
      description='Tools to design guides for diagnostics',
      long_description=LONG_DESCRIPTION,
      long_description_content_type='text/markdown',
      author='Hayden Metsky',
      author_email='hmetsky@broadinstitute.org',
      maintainer='Priya P. Pillai',
      maintainer_email='ppillai@broadinstitute.org',
      url="https://github.com/broadinstitute/adapt",
      version=adapt.__version__,
      packages=find_packages(),
      package_data={
        "adapt": ["models/*/*/*/assets.extra/*", "models/*/*/*/variables/*", "models/*/*/*/*", "models/*/*/*"],
      },
      install_requires=['numpy>=1.16.0,<1.19.0', 'scipy==1.4.1', 'tensorflow==2.3.2'],
      extras_require={
        'AWS': ['boto3>=1.14.54', 'botocore>=1.17.54']
      },
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
      ],
      scripts=[
        'bin/design.py',
        'bin/design_naively.py',
        'bin/analyze_coverage.py',
        'bin/pick_test_targets.py'
      ])



"""Structure(s) and functions for working with alignments of sequences.
"""

from collections import defaultdict
import logging
import statistics

import numpy as np
from math import log2

from adapt.utils import guide
from adapt.utils import lsh
from adapt.utils import predict_activity

__author__ = 'Hayden Metsky <hayden@mit.edu>'

logger = logging.getLogger(__name__)


class SequenceList:
    """Immutable collection of sequences.
    """

    def __init__(self, seqs):
        """
        Args:
            seqs: list of str, where seqs[i] is the i'th sequence
        """
        self.seqs = seqs
        self.num_sequences = len(seqs)

    def make_list_of_seqs(self, seqs_to_consider=None, include_idx=False,
                          remove_gaps=False):
        """Construct list of sequences.

        Args:
            seqs_to_consider: collection of indices of sequences to use (if None,
                use all)
            include_idx: instead of a list of str giving the sequences in
                the alignment, return a list of tuples (seq, idx) where seq
                is a str giving a sequence and idx is the index in the
                list
            remove_gaps: if True, remove gaps ('-') from the returned
                sequences; note that there should not be gaps because,
                in general, this should represent unaligned sequences

        Returns:
            list of str giving the sequences (or, list of
            tuples if include_idx is True)
        """
        if seqs_to_consider is None:
            seqs_to_consider = range(self.num_sequences)

        def seq_str(i):
            # Get the str of the sequence with index i
            s = self.seqs[i]
            if remove_gaps:
                s = s.replace('-', '')
            return s

        if include_idx:
            return [(seq_str(i), i) for i in seqs_to_consider]
        else:
            return [seq_str(i) for i in seqs_to_consider]


class Alignment(SequenceList):
    """Immutable collection of sequences that have been aligned.

    This stores sequences in column-major order, which should make it more
    efficient to extract pieces of the alignment by position and to generate
    consensus sequences.
    """

    def __init__(self, seqs):
        """
        Args:
            seqs: list of str representing an alignment in column-major order
                (i.e., seqs[i] is a string giving the bases in the sequences
                at the i'th position of the alignment; it is not the i'th sequence)
        """
        self.seq_length = len(seqs)
        self.num_sequences = len(seqs[0])
        for s in seqs:
            assert len(s) == self.num_sequences

        self.seqs = seqs

        # Memoize information missing data at each position
        self._frac_missing = None
        self._median_missing = None

    def extract_range(self, pos_start, pos_end):
        """Extract range of positions from alignment.

        Args:
            pos_start: start position of extraction (inclusive)
            pos_end: end position of extraction (exclusive)

        Returns:
            object of type Alignment including only the specified range
        """
        return Alignment(self.seqs[pos_start:pos_end])

    def _compute_frac_missing(self):
        """Compute fraction of sequences with missing data at each position.
        """
        self._frac_missing = [0 for _ in range(self.seq_length)]
        for j in range(self.seq_length):
            num_n = sum(1 for i in range(self.num_sequences)
                        if self.seqs[j][i] == 'N')
            self._frac_missing[j] = float(num_n) / self.num_sequences

    def median_sequences_with_missing_data(self):
        """Compute the median fraction of sequences with missing data, across
        positions.

        Returns:
            median fraction of sequences that have 'N', taken across the
            positions in the alignment
        """
        if self._median_missing is None:
            if self._frac_missing is None:
                self._compute_frac_missing()
            self._median_missing = statistics.median(self._frac_missing)
        return self._median_missing

    def frac_missing_at_pos(self, pos):
        """Find fraction of sequences with missing data at position.

        Args:
            pos: position in the alignment at which to return the fraction
                of sequences with 'N'

        Returns:
            fraction of sequences at pos with 'N'
        """
        if self._frac_missing is None:
            self._compute_frac_missing()
        return self._frac_missing[pos]

    def seqs_with_gap(self, seqs_to_consider=None):
        """Determine sequences in the alignment that have a gap.

        Args:
            seqs_to_consider: only look within seqs_to_consider for
                sequences with a gap; if None, then look in all
                sequences

        Returns:
            set of indices of sequences that contain a gap
        """
        if seqs_to_consider is None:
            seqs_to_consider = range(self.num_sequences)

        has_gap = set()
        for j in range(self.seq_length):
            has_gap.update(i for i in seqs_to_consider if self.seqs[j][i] == '-')
        return has_gap

    def seqs_with_required_flanking(self, guide_start, guide_length,
            required_flanking_seqs, seqs_to_consider=None):
        """Determine sequences in the alignment with required flanking sequence.

        If no flanking sequences are required, this says that all sequences
        have the required flanking sequence.

        Args:
            guide_start: start position of the guide in the alignment
            guide_length: length of the guide
            required_flanking_seqs: tuple (s5, s3) that specifies sequences
                on the 5' (left; s5) end and 3' (right; s3) end flanking
                the guide (in the target, not the guide) that must be
                required for a guide to bind; if either is None, no
                flanking sequence is required for that end
            seqs_to_consider: only look within seqs_to_consider for
                sequences with the flanking sequences; if None, then
                look in all sequences

        Returns:
            set of indices of sequences that have the required flanking
            sequences
        """
        if seqs_to_consider is None:
            seqs_to_consider = set(range(self.num_sequences))

        def seqs_that_equal(query, start_pos, end_pos):
            # Return set of sequence indices from seqs_to_consider that,
            # in the range [start_pos, end_pos) do not contain a gap and
            # are equal to query
            if start_pos < 0 or end_pos > self.seq_length:
                return set()
            query_matches = set()
            for i in seqs_to_consider:
                s = ''.join(self.seqs[j][i] for j in range(start_pos, end_pos))
                if '-' not in s and guide.query_target_eq(query, s):
                    query_matches.add(i)
            return query_matches

        required_flanking5, required_flanking3 = required_flanking_seqs

        seqs_with_required_flanking = set(seqs_to_consider)
        if required_flanking5 is not None and len(required_flanking5) > 0:
            query = required_flanking5
            seqs_with_required_flanking &= seqs_that_equal(
                required_flanking5, guide_start - len(required_flanking5),
                guide_start)
        if required_flanking3 is not None and len(required_flanking3) > 0:
            seqs_with_required_flanking &= seqs_that_equal(
                required_flanking3, guide_start + guide_length,
                guide_start + guide_length + len(required_flanking3))
        return seqs_with_required_flanking

    def construct_guide(self, start, guide_length, seqs_to_consider, mismatches,
            allow_gu_pairs, guide_clusterer, num_needed=None,
            missing_threshold=1, guide_is_suitable_fn=None,
            required_flanking_seqs=(None, None),
            predictor=None, stop_early=True):
        """Construct a single guide to target a set of sequences in the alignment.

        This constructs a guide to target sequence within the range [start,
        start+guide_length]. It only considers the sequences with indices given in
        seqs_to_consider.

        Args:
            start: start position in alignment at which to target
            guide_length: length of the guide
            seqs_to_consider: dict mapping universe group ID to collection of
                indices to use when constructing the guide
            mismatches: threshold on number of mismatches for determining whether
                a guide would hybridize to a target sequence
            allow_gu_pairs: if True, tolerate G-U base pairs between a guide
                and target when computing whether a guide binds
            guide_clusterer: object of SequenceClusterer to use for clustering
                potential guide sequences; it must have been initialized with
                a family suitable for guides of length guide_length; if None,
                then don't cluster, and instead draw a consensus from all
                the sequences
            num_needed: dict mapping universe group ID to the number of sequences
                from the group that are left to cover in order to achieve
                a desired coverage; these are used to help construct a
                guide
            missing_threshold: do not construct a guide if the fraction of
                sequences with missing data, at any position in the target
                range, exceeds this threshold
            guide_is_suitable_fn: if set, a function f(x) such that this
                will only construct a guide x for which f(x) is True
            required_flanking_seqs: tuple (s5, s3) that specifies sequences
                on the 5' (left; s5) end and 3' (right; s3) end flanking
                the guide (in the target, not the guide) that must be
                present for a guide to bind; if either is None, no
                flanking sequence is required for that end
            predictor: if set, a adapt.utils.predict_activity.Predictor
                object. If None, do not predict activity.
            stop_early: if True, impose early stopping criteria while iterating
                over clusters to improve runtime

        Returns:
            tuple (x, y) where:
                x is the sequence of the constructed guide
                y is a set of indices of sequences (a subset of
                    values in seqs_to_consider) to which the guide x will
                    hybridize
            (Note that it is possible that x binds to no sequences and that
            y will be empty.)
        """
        # TODO: There are several optimizations that can be made to
        # this function that take advantage of G-U pairing in order
        # to lower the number of guides that need to be designed.
        # Two are:
        #  1) The function SequenceClusterer.cluster(..), which is
        #     used here, can cluster accounting for G-U pairing (e.g.,
        #     such that 'A' hashes to 'G' and 'C' hashes to 'T',
        #     so that similar guide sequences hash to the same
        #     value tolerating G-U similarity).
        #  2) Instead of taking a consensus sequence across a
        #     cluster, this can take a pseudo-consensus that
        #     accounts for G-U pairing (e.g., in which 'A' is
        #     treated as 'G' and 'C' is treated as 'T' for
        #     the purposes of generating a consensus sequence).

        assert start + guide_length <= self.seq_length
        assert len(seqs_to_consider) > 0

        for pos in range(start, start + guide_length):
            if self.frac_missing_at_pos(pos) > missing_threshold:
                raise CannotConstructGuideError(("Too much missing data at "
                    "a position in the target range"))

        aln_for_guide = self.extract_range(start, start + guide_length)

        if predictor is not None:
            # Extract the target sequences, including context to use with
            # prediction
            if (start - predictor.context_nt < 0 or
                    start + guide_length + predictor.context_nt > self.seq_length):
                raise CannotConstructGuideError(("The context needed for "
                    "the target to predict activity falls outside the "
                    "range of the alignment at this position"))
            aln_for_guide_with_context = self.extract_range(
                    start - predictor.context_nt,
                    start + guide_length + predictor.context_nt)

        # Before modifying seqs_to_consider, make a copy of it
        seqs_to_consider_cp = {}
        for group_id in seqs_to_consider.keys():
            seqs_to_consider_cp[group_id] = set(seqs_to_consider[group_id])
        seqs_to_consider = seqs_to_consider_cp

        all_seqs_to_consider = set.union(*seqs_to_consider.values())

        # Ignore any sequences in the alignment that have a gap in
        # this region
        if predictor is not None:
            seqs_with_gap = set(aln_for_guide_with_context.seqs_with_gap(all_seqs_to_consider))
        else:
            seqs_with_gap = set(aln_for_guide.seqs_with_gap(all_seqs_to_consider))
        for group_id in seqs_to_consider.keys():
            seqs_to_consider[group_id].difference_update(seqs_with_gap)
        all_seqs_to_consider = set.union(*seqs_to_consider.values())

        # Only consider sequences in the alignment that have the
        # required flanking sequence(s)
        seqs_with_flanking = self.seqs_with_required_flanking(
            start, guide_length, required_flanking_seqs,
            seqs_to_consider=all_seqs_to_consider)
        for group_id in seqs_to_consider.keys():
            seqs_to_consider[group_id].intersection_update(seqs_with_flanking)
        all_seqs_to_consider = set.union(*seqs_to_consider.values())

        # If every sequence in this region has a gap or does not contain
        # required flanking sequence, then there are none left to consider
        if len(all_seqs_to_consider) == 0:
            raise CannotConstructGuideError(("All sequences in region have "
                "a gap and/or do not contain required flanking sequences"))

        seq_rows = aln_for_guide.make_list_of_seqs(all_seqs_to_consider,
            include_idx=True)
        if predictor is not None:
            seq_rows_with_context = aln_for_guide_with_context.make_list_of_seqs(
                    all_seqs_to_consider, include_idx=True)

        if predictor is not None:
            # Memoize activity evaluations
            pair_eval = {}
        def determine_binding_and_active_seqs(gd_sequence):
            binding_seqs = set()
            num_bound = 0
            if predictor is not None:
                num_passed_predict_active = 0
                # Determine what calls to make to
                # predictor.determine_highly_active(); it
                # is best to batch these
                pairs_to_eval = []
                for i, (seq, seq_idx) in enumerate(seq_rows):
                    if guide.guide_binds(gd_sequence, seq, mismatches,
                            allow_gu_pairs):
                        seq_with_context, _ = seq_rows_with_context[i]
                        pair = (seq_with_context, gd_sequence)
                        pairs_to_eval += [pair]
                # Evaluate activity
                evals = predictor.determine_highly_active(start, pairs_to_eval)
                for pair, y in zip(pairs_to_eval, evals):
                    pair_eval[pair] = y
                # Fill in binding_seqs
                for i, (seq, seq_idx) in enumerate(seq_rows):
                    if guide.guide_binds(gd_sequence, seq, mismatches,
                            allow_gu_pairs):
                        num_bound += 1
                        seq_with_context, _ = seq_rows_with_context[i]
                        pair = (seq_with_context, gd_sequence)
                        if pair_eval[pair]:
                            num_passed_predict_active += 1
                            binding_seqs.add(seq_idx)
            else:
                num_passed_predict_active = None
                for i, (seq, seq_idx) in enumerate(seq_rows):
                    if guide.guide_binds(gd_sequence, seq, mismatches,
                            allow_gu_pairs):
                        num_bound += 1
                        binding_seqs.add(seq_idx)
            return binding_seqs, num_bound, num_passed_predict_active

        # Define a score function (higher is better) for a collection of
        # sequences covered by a guide
        if num_needed is not None:
            # This is the number of sequences it contains that are needed to
            # achieve the partial cover; we can compute this by summing over
            # the number of needed sequences it contains, taken across the
            # groups in the universe
            # Memoize the scores because this computation might be expensive
            seq_idxs_scores = {}
            def seq_idxs_score(seq_idxs):
                seq_idxs = set(seq_idxs)
                tc = tuple(seq_idxs)
                if tc in seq_idxs_scores:
                    return seq_idxs_scores[tc]
                score = 0
                for group_id, needed in num_needed.items():
                    contained_in_seq_idxs = seq_idxs & seqs_to_consider[group_id]
                    score += min(needed, len(contained_in_seq_idxs))
                seq_idxs_scores[tc] = score
                return score
        else:
            # Score by the number of sequences it contains
            def seq_idxs_score(seq_idxs):
                return len(seq_idxs)

        # First construct the optimal guide to cover the sequences. This would be
        # a string x that maximizes the number of sequences s_i such that x and
        # s_i are equal to within 'mismatches' mismatches; it's called the "max
        # close string" or "close to most strings" problem. For simplicity, let's
        # do the following: cluster the sequences (just the portion with
        # potential guides) with LSH, choose a guide for each cluster to be
        # the consensus of that cluster, and choose the one (across clusters)
        # that has the highest score
        if guide_clusterer is None:
            # Don't cluster; instead, draw the consensus from all the
            # sequences. Effectively, treat it as there being just one
            # cluster, which consists of all sequences to consider
            clusters_ordered = [all_seqs_to_consider]
        else:
            # Cluster the sequences
            clusters = guide_clusterer.cluster(seq_rows)

            # Sort the clusters by score, from highest to lowest
            # Here, the score is determined by the sequences in the cluster
            clusters_ordered = sorted(clusters, key=seq_idxs_score, reverse=True)

        best_gd = None
        best_gd_binding_seqs = None
        best_gd_score = 0
        stopped_early = False
        for cluster_idxs in clusters_ordered:
            if stop_early and best_gd_score > seq_idxs_score(cluster_idxs):
                # The guide from this cluster is unlikely to exceed the current
                # score; stop early
                stopped_early = True
                break

            gd = aln_for_guide.determine_consensus_sequence(
                cluster_idxs)
            if 'N' in gd:
                # Skip this; all sequences at a position in this cluster
                # are 'N'
                continue
            if guide_is_suitable_fn is not None:
                if guide_is_suitable_fn(gd) is False:
                    # Skip this cluster
                    continue
            # Determine the sequences that are bound by this guide (and
            # where it is 'active', if predictor is set)
            binding_seqs, num_bound, num_passed_predict_active = \
                    determine_binding_and_active_seqs(gd)
            score = seq_idxs_score(binding_seqs)
            if score > best_gd_score:
                best_gd = gd
                best_gd_binding_seqs = binding_seqs
                best_gd_score = score

            # Impose an early stopping criterion if predictor is
            # used, because using it is slow
            if predictor is not None and stop_early:
                if (num_bound >= 0.5*len(cluster_idxs) and
                        num_passed_predict_active < 0.1*len(cluster_idxs)):
                    # gd binds (according to guide.guide_binds()) many
                    # sequences, but is not predicted to be active against
                    # many; it is likely that this region is poor according to
                    # the predictor (e.g., due to sequence composition). Rather
                    # than trying other clusters at this site, just skip it.
                    # Note that this is just a heuristic; it can help runtime
                    # when the predictor is used, but is not needed and may
                    # hurt the optimality of the solution
                    stopped_early = True
                    break
        gd = best_gd
        binding_seqs = best_gd_binding_seqs

        # It's possible that the consensus sequence (guide) of no cluster
        # binds to any of the sequences. In this case, simply go through all
        # sequences and find the first that has no ambiguity and is suitable
        # and active, and make this the guide
        if gd is None and not stopped_early:
            for i, (s, idx) in enumerate(seq_rows):
                if not set(s).issubset(set(['A', 'C', 'G', 'T'])):
                    # s has ambiguity; skip it
                    continue
                if (guide_is_suitable_fn is not None and
                        guide_is_suitable_fn(s) is False):
                    # Skip s, which is not suitable
                    continue
                if predictor is not None:
                    s_with_context, _ = seq_rows_with_context[i]
                    if not predictor.determine_highly_active(start, [(s_with_context, s)])[0]:
                        # s is not active against itself; skip it
                        continue
                # s has no ambiguity and is a suitable guide; use it
                gd = s
                binding_seqs, _, _ = determine_binding_and_active_seqs(gd)
                break

        if gd is None:
            raise CannotConstructGuideError(("No guides are suitable or "
                "active"))

        return (gd, binding_seqs)

    def make_list_of_seqs(self, seqs_to_consider=None, include_idx=False,
                          remove_gaps=False):
        """Construct list of sequences from the alignment.

        Args:
            seqs_to_consider: collection of indices of sequences to use (if None,
                use all)
            include_idx: instead of a list of str giving the sequences in
                the alignment, return a list of tuples (seq, idx) where seq
                is a str giving a sequence and idx is the index in the
                alignment
            remove_gaps: if True, remove gaps ('-') from the returned
                sequences (i.e., so the sequences are as they appeared
                prior to alignment); in this case, the returned sequences
                may be of differing lengths

        Returns:
            list of str giving the sequences in the alignment (or, list of
            tuples if include_idx is True)
        """
        if seqs_to_consider is None:
            seqs_to_consider = range(self.num_sequences)

        def seq_str(i):
            # Construct the str of the sequence with index i
            s = ''.join(self.seqs[j][i] for j in range(self.seq_length))
            if remove_gaps:
                s = s.replace('-', '')
            return s

        if include_idx:
            return [(seq_str(i), i) for i in seqs_to_consider]
        else:
            return [seq_str(i) for i in seqs_to_consider]

    def determine_consensus_sequence(self, seqs_to_consider=None):
        """Determine consensus sequence from the alignment.

        At each position, the consensus is the most common allele even if it is
        not the majority; the consensus will not be an ambiguity code (except N
        if all bases are N). Ties are broken arbitrarily but deterministically.

        This ignores 'N' bases at each position, determining the consensus from
        the other (non-'N') bases. An 'N' will only be output in the consensus
        if all bases at a position are 'N'.

        Args:
            seqs_to_consider: collection of indices of sequences to use (if None,
                use all)

        Returns:
            str representing the consensus of the alignment
        """
        if seqs_to_consider is None:
            seqs_to_consider = range(self.num_sequences)

        consensus = ''
        for i in range(self.seq_length):
            counts = {'A': 0, 'T': 0, 'C': 0, 'G': 0}
            for b in [self.seqs[i][j] for j in seqs_to_consider]:
                if b in counts:
                    counts[b] += 1
                elif b == 'N':
                    # skip N
                    continue
                elif b in guide.FASTA_CODES:
                    for c in guide.FASTA_CODES[b]:
                        counts[c] += 1.0 / len(guide.FASTA_CODES[b])
                else:
                    raise ValueError("Unknown base call %s" % b)
            counts_sorted = sorted(list(counts.items()))
            max_base = max(counts_sorted, key=lambda x: x[1])[0]
            if counts[max_base] == 0:
                consensus += 'N'
            else:
                consensus += max_base

        return consensus

    def determine_most_common_sequence(self, seqs_to_consider=None,
            skip_ambiguity=False):
        """Determine the most common of the sequences from the alignment.

        If we imagine each sequence in the alignment as existing some
        number of times in the alignment, this effectively picks the
        mode.

        Ties are broken arbitrarily but deterministically.

        Args:
            seqs_to_consider: collection of indices of sequences to use (if
                None, use all)
            skip_ambiguity: if True, ignore any sequences that contain
                an ambiguity code (i.e., only count sequences where all
                bases are 'A', 'T', 'C', or 'G')

        Returns:
            str representing the mode of the sequences (or None if there
            are no suitable strings)
        """
        if seqs_to_consider is None:
            seqs_to_consider = range(self.num_sequences)

        allowed_chars = ['A', 'T', 'C', 'G', '-']

        # Convert all sequences into strings, and count the number of each
        seqs_str = self.make_list_of_seqs(seqs_to_consider=seqs_to_consider)
        seq_count = defaultdict(int)
        for seq_str in seqs_str:
            if skip_ambiguity:
                # Skip over this sequence if it contains any base that is
                # ambiguous
                skip = False
                for i in range(len(seq_str)):
                    if seq_str[i] not in allowed_chars:
                        skip = True
                        break
                if skip:
                    continue

            seq_count[seq_str] += 1

        if len(seq_count) == 0:
            # There are no suitable strings (e.g., all contain ambiguity)
            return None

        # Find the most common sequence (sort to break ties deterministically)
        counts_sorted = sorted(list(seq_count.items()))
        max_seq_str = max(counts_sorted, key=lambda x: x[1])[0]
        return max_seq_str

    def determine_representative_guides(self, start, guide_length,
            seqs_to_consider, guide_clusterer, missing_threshold=1,
            guide_is_suitable_fn=None,
            required_flanking_seqs=(None, None)):
        """Construct a set of guides representative of the target sequences.

        This is similar to construct_guide(), except returns a set of
        representative sequences rather than a single best one.

        Args:
            start: start position in alignment at which to target
            guide_length: length of the representative sequence (guide)
            seqs_to_consider: dict mapping universe group ID to collection of
                indices to use when constructing the representative sequences
            guide_clusterer: object of SequenceClusterer to use for clustering
                potential guide sequences; it must have been initialized with
                a family suitable for guides of length guide_length; if None,
                then don't cluster, and instead draw a consensus from all
                the sequences
            missing_threshold: do not construct representative sequences if
                the fraction of sequences with missing data, at any position
                in the target range, exceeds this threshold
            guide_is_suitable_fn: if set, a function f(x) such that this
                will only construct a guide x for which f(x) is True
            required_flanking_seqs: tuple (s5, s3) that specifies sequences
                on the 5' (left; s5) end and 3' (right; s3) end flanking
                the guide (in the target, not the guide) that must be
                present for a guide to bind; if either is None, no
                flanking sequence is required for that end

        Returns:
            set of representative sequences
        """
        assert start + guide_length <= self.seq_length
        assert len(seqs_to_consider) > 0

        for pos in range(start, start + guide_length):
            if self.frac_missing_at_pos(pos) > missing_threshold:
                raise CannotConstructGuideError(("Too much missing data at "
                    "a position in the target range"))

        aln_for_guide = self.extract_range(start, start + guide_length)

        # Before modifying seqs_to_consider, make a copy of it
        seqs_to_consider_cp = {}
        for group_id in seqs_to_consider.keys():
            seqs_to_consider_cp[group_id] = set(seqs_to_consider[group_id])
        seqs_to_consider = seqs_to_consider_cp

        all_seqs_to_consider = set.union(*seqs_to_consider.values())

        # Ignore any sequences in the alignment that have a gap in
        # this region
        seqs_with_gap = set(aln_for_guide.seqs_with_gap(all_seqs_to_consider))
        for group_id in seqs_to_consider.keys():
            seqs_to_consider[group_id].difference_update(seqs_with_gap)
        all_seqs_to_consider = set.union(*seqs_to_consider.values())

        # Only consider sequences in the alignment that have the
        # required flanking sequence(s)
        seqs_with_flanking = self.seqs_with_required_flanking(
            start, guide_length, required_flanking_seqs,
            seqs_to_consider=all_seqs_to_consider)
        for group_id in seqs_to_consider.keys():
            seqs_to_consider[group_id].intersection_update(seqs_with_flanking)
        all_seqs_to_consider = set.union(*seqs_to_consider.values())

        # If every sequence in this region has a gap or does not contain
        # required flanking sequence, then there are none left to consider
        if len(all_seqs_to_consider) == 0:
            raise CannotConstructGuideError(("All sequences in region have "
                "a gap and/or do not contain required flanking sequences"))

        seq_rows = aln_for_guide.make_list_of_seqs(all_seqs_to_consider,
            include_idx=True)

        # Cluster the sequences
        clusters = guide_clusterer.cluster(seq_rows)

        # Take the consensus of each cluster to be the representative
        representatives = set()
        for cluster_idxs in clusters:
            gd = aln_for_guide.determine_consensus_sequence(
                cluster_idxs)
            if 'N' in gd:
                # Skip this; all sequences at a position in this cluster
                # are 'N'
                continue
            if guide_is_suitable_fn is not None:
                if guide_is_suitable_fn(gd) is False:
                    # Skip this cluster; it is not suitable (e.g., may
                    # not be specific)
                    continue
            representatives.add(gd)
        return representatives

    def compute_activity(self, start, gd_sequence, predictor, mutator=None):
        """Compute activity between a guide sequence and every target sequence
        in the alignment.

        This only considers the region of the alignment within the range
        [start, start + len(gd_sequence)), along with sequence context.

        Args:
            start: start position in alignment at which to target
            gd_sequence: str representing guide sequence
            predictor: a adapt.utils.predict_activity.Predictor object
            mutator: a adapt.utils.mutate.Mutator object

        Returns:
            numpy array x where x[i] gives the predicted activity between
            gd_sequence and the sequence in the alignment at index i. If
            mutator is not None, x[i] gives the predicted activity after the
            mutations specified in the mutator.
        """
        guide_length = len(gd_sequence)
        assert start + guide_length <= self.seq_length

        if isinstance(predictor, predict_activity.SimpleBinaryPredictor):
            if mutator:
                # Extract the target sequences, including context to use with
                # prediction (i.e. the flanking sequences, if they exist)
                left_context = 0
                right_context = 0
                if predictor:
                    if predictor.required_flanking_seqs[0]:
                        left_context = len(predictor.required_flanking_seqs[0])
                    if predictor.required_flanking_seqs[1]:
                        right_context = len(predictor.required_flanking_seqs[1])

                aln_for_guide_with_context = self.extract_range(
                        start - left_context,
                        start + guide_length + right_context)
                seq_rows_with_context = aln_for_guide_with_context.make_list_of_seqs()

                # Start array of predicted activities; make this all 0s so that
                # sequences for which activities are not computed (e.g., gap)
                # have activity=0
                activities = np.zeros(self.num_sequences)
                for i, seq_with_context in enumerate(seq_rows_with_context):
                    activity = mutator.compute_mutated_activity(predictor,
                                                                seq_with_context,
                                                                gd_sequence,
                                                                start=start)
                    activities[i] = activity
                return activities
            # Do not use a model; just predict binary activity (1 or 0)
            # based on distance between guide and targets
            return predictor.compute_activity(start, gd_sequence, self)

        # Extract the target sequences, including context to use with
        # prediction
        if (start - predictor.context_nt < 0 or
                start + guide_length + predictor.context_nt > self.seq_length):
            raise CannotConstructGuideError(("The context needed for "
                "the target to predict activity falls outside the "
                "range of the alignment at this position"))
        aln_for_guide_with_context = self.extract_range(
                start - predictor.context_nt,
                start + guide_length + predictor.context_nt)

        # Ignore any sequences in the alignment that have a gap in
        # this region
        all_seqs_to_consider = set(range(self.num_sequences))
        seqs_with_gap = set(aln_for_guide_with_context.seqs_with_gap(all_seqs_to_consider))
        seqs_to_consider = all_seqs_to_consider.difference(seqs_with_gap)

        seq_rows_with_context = aln_for_guide_with_context.make_list_of_seqs(
                seqs_to_consider, include_idx=True)

        # Start array of predicted activities; make this all 0s so that
        # sequences for which activities are not computed (e.g., gap)
        # have activity=0
        activities = np.zeros(self.num_sequences)

        # Determine what calls to make to predictor.compute_activity(); it
        # is best to batch these
        pairs_to_eval = []
        pairs_to_eval_seq_idx = []
        evals = []

        if mutator:
            for seq_with_context, seq_idx in seq_rows_with_context:
                activity = mutator.compute_mutated_activity(predictor,
                                                            seq_with_context,
                                                            gd_sequence,
                                                            start=start)
                evals.append(activity)
                pairs_to_eval_seq_idx.append(seq_idx)
        else:
            for seq_with_context, seq_idx in seq_rows_with_context:
                pair = (seq_with_context, gd_sequence)
                pairs_to_eval.append(pair)
                pairs_to_eval_seq_idx.append(seq_idx)
            # Evaluate activity
            evals = predictor.compute_activity(start, pairs_to_eval)

        for activity, seq_idx in zip(evals, pairs_to_eval_seq_idx):
            # Fill in the activity for seq_idx
            activities[seq_idx] = activity

        return activities

    def sequences_bound_by_guide(self, gd_seq, gd_start, mismatches,
            allow_gu_pairs, required_flanking_seqs=(None, None)):
        """Determine the sequences to which a guide hybridizes.

        Args:
            gd_seq: sequence of the guide
            gd_start: start position of the guide in the alignment
            mismatches: threshold on number of mismatches for determining whether
                a guide would hybridize to a target sequence
            allow_gu_pairs: if True, tolerate G-U base pairs between a
                guide and target when computing whether a guide binds
            required_flanking_seqs: tuple (s5, s3) that specifies sequences
                on the 5' (left; s5) end and 3' (right; s3) end flanking
                the guide (in the target, not the guide) that must be
                present for a guide to bind; if either is None, no
                flanking sequence is required for that end

        Returns:
            collection of indices of sequences to which the guide will
            hybridize
        """
        assert gd_start + len(gd_seq) <= self.seq_length

        aln_for_guide = self.extract_range(gd_start, gd_start + len(gd_seq))
        seq_rows = aln_for_guide.make_list_of_seqs(include_idx=True)

        seqs_with_flanking = self.seqs_with_required_flanking(
            gd_start, len(gd_seq), required_flanking_seqs)

        binding_seqs = []
        for seq, seq_idx in seq_rows:
            if (seq_idx in seqs_with_flanking and
                    guide.guide_binds(gd_seq, seq, mismatches, allow_gu_pairs)):
                binding_seqs += [seq_idx]
        return binding_seqs

    def position_entropy(self):
        """Determine the entropy at each position in the alignment.

        Returns:
            list with the entropy of position i listed at index i
        """

        position_entropy = []
        for i in range(self.seq_length):
            counts = {'A': 0, 'T': 0, 'C': 0, 'G': 0, '-': 0}
            for b in [self.seqs[i][j] for j in range(self.num_sequences)]:
                if b in counts:
                    counts[b] += 1
                elif b in guide.FASTA_CODES:
                    for c in guide.FASTA_CODES[b]:
                        counts[c] += 1.0 / len(guide.FASTA_CODES[b])
                else:
                    raise ValueError("Unknown base call %s" % b)

            # Calculate entropy
            probabilities = [counts[base]/self.num_sequences for base in counts]
            this_position_entropy = sum([-p*log2(p) for p in probabilities if p > 0])

            position_entropy.append(this_position_entropy)

        return position_entropy

    def base_percentages(self):
        """Determines the percentage of each base pair in the alignment.

        Returns:
            dictionary of base pair to its percentage in the alignment
        """
        counts = {'A': 0, 'T': 0, 'C': 0, 'G': 0}
        for i in range(self.seq_length):
            for b in [self.seqs[i][j] for j in range(self.num_sequences)]:
                if b in counts:
                    counts[b] += 1
                elif b in guide.FASTA_CODES:
                    for c in guide.FASTA_CODES[b]:
                        counts[c] += 1.0 / len(guide.FASTA_CODES[b])
                elif b != '-':
                    raise ValueError("Unknown base call %s" % b)
        total = sum(counts.values())
        for base in counts:
            counts[base] /= total
        return counts

    @staticmethod
    def from_list_of_seqs(seqs):
        """Construct a Alignment from aligned list of sequences.

        If seqs is stored in row-major order, this converts to column-major order
        and creates a Alignment object.

        Args:
            seqs: list of str, all the same length, representing an alignment;
                seqs[i] is the i'th sequence

        Returns:
            object of type Alignment
        """
        num_sequences = len(seqs)

        if num_sequences == 0:
            raise Exception(("Trying to construct an alignment consisting "
                "of 0 sequences"))

        seq_length = len(seqs[0])
        for s in seqs:
            if len(s) != seq_length:
                raise ValueError("Sequences must be the same length")

        seqs_col = ['' for _ in range(seq_length)]
        for j in range(seq_length):
            seqs_col[j] = ''.join(seqs[i][j] for i in range(num_sequences))

        return Alignment(seqs_col)


class CannotConstructGuideError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


class SequenceClusterer:
    """Supports clustering sequences using LSH.

    This is useful for approximating an optimal guide over a collection of
    sequences.
    """

    def __init__(self, family, k=10):
        """
        Args:
            family: object of hash family from which to draw hash functions
            k: number of hash functions to draw from a family of
                hash functions for amplification; each hash function is then
                the concatenation (h_1, h_2, ..., h_k)
        """
        self.hash_concat = lsh.HashConcatenation(family, k)

    def cluster(self, seqs_with_idx):
        """Generate clusters of given sequences.

        Args:
            seqs_with_idx: collection of tuples (seq, idx) where seq
                is a str giving a sequence and idx is an identifier
                of the sequence (e.g., index in an alignment); all idx
                in the collection must be unique

        Returns:
            collection {S_i} where each S_i is a set of idx from
            seqs_with_idx that form a cluster; the S_i's are disjoint and
            their union is all the idx
        """
        def g(seq):
            # Hash a sequence; since the output of self.hash_concat is a
            # tuple of strings, we can join it into one string
            return ''.join(self.hash_concat.g(seq))

        d = defaultdict(set)
        for seq, idx in seqs_with_idx:
            d[g(seq)].add(idx)
        return d.values()

    def largest_cluster(self, seqs_with_idx):
        """Find sequences that form the largest cluster.

        Args:
            seqs_with_idx: collection of tuples (seq, idx) where seq
                is a str giving a sequence and idx is an identifier
                of the sequence (e.g., index in an alignment); all idx
                in the collection must be unique

        Returns:
            set of idx from seqs_with_idx whose sequences form the largest
            cluster
        """
        clusters = self.cluster(seqs_with_idx)
        return max(clusters, key=lambda c: len(c))




"""Methods for searching for amplicons and guides within them.

This combines primer_search, to search for amplicons bound by primers,
and guide_search, to search for guides within those amplicons. We
term the combination of these a "target".

These methods score possible targets and search for the top N of them.
"""

import heapq
import itertools
import logging
import math
import os
import numpy as np

from adapt import guide_search
from adapt import primer_search

__author__ = 'Hayden Metsky <hayden@mit.edu>'

logger = logging.getLogger(__name__)


class TargetSearcher:
    """Methods to search for targets over a genome."""

    def __init__(self, ps, gs, obj_type='min',
            max_primers_at_site=None,
            max_target_length=None, obj_weights=None,
            only_account_for_amplified_seqs=False,
            halt_early=False, obj_value_shift=None,
            mutator=None):
        """
        Args:
            ps: PrimerSearcher object
            gs: GuideSearcher object
            obj_type: 'min' or 'max' indicating whether the objective
                should be considered to be a minimization or maximization
                problem
            max_primers_at_site: only allow amplicons in which each
                end has at most this number of primers; or None for
                no limit
            max_target_length: only allow amplicons whose length is at
                most this; or None for no limit
            obj_weights: a tuple giving weights in the target objective
                function for penalties on number of primers and amplicon
                length, relative to the guide objective value. Tuple
                (A, B) where A weighs total number of primers and B weighs
                amplicon length
            only_account_for_amplified_seqs: design guides so as to account for
                only sequences bound by primers (e.g., when obj_type is 'min',
                cover only sequences bound by the primers), rather than
                all sequences; must be False if obj_type is 'max'
            halt_early: if True, stop as soon as there are the desired number
                of targets found, even if this does not complete the
                search over the whole genome (i.e., the targets meet the
                constraints but may not be optimal)
            obj_value_shift: amount by which to shift objective values
                before they are reported. This is only intended to avoid
                confusion about reported objective values -- e.g., if
                values for some targets would be positive and others
                negative, we can shift all values upward so they are all
                positive so that there is not confusion about differences
                between positive/negative. If None (default), defaults are
                set below based on obj_type.
            mutator: a adapt.utils.mutate Mutator object. If None (default),
                do not predict activity after mutations.
        """
        self.ps = ps
        self.gs = gs

        if obj_type not in ['min', 'max']:
            raise ValueError(("obj_type must be 'min' or 'max'"))
        self.obj_type = obj_type

        self.max_primers_at_site = max_primers_at_site
        self.max_target_length = max_target_length

        if obj_weights is None:
            obj_weights = (0.50, 0.25)
        self.obj_weight_primers = obj_weights[0]
        self.obj_weight_length = obj_weights[1]

        if (only_account_for_amplified_seqs and
                isinstance(self.gs, guide_search.GuideSearcherMaximizeActivity)):
            # GuideSearcherMaximizeActivity only considers all sequences --
            # this is mostly for technical implementation reasons
            raise ValueError(("When maximizing activity, "
                "only_account_for_amplified_seqs must be False"))

        self.only_account_for_amplified_seqs = only_account_for_amplified_seqs

        self.halt_early = halt_early

        if obj_value_shift is None:
            if obj_type == 'min':
                # Values for all design options should all be positive, so
                # no need to shift
                self.obj_value_shift = 0
            elif obj_type == 'max':
                # Values for some design options may be positive and others
                # negative, so shift all up so that most are positive
                self.obj_value_shift = 4.0

        self.mutator = mutator

    def _find_primer_pairs(self):
        """Find suitable primer pairs using self.ps.

        Yields:
            tuple (p_i, p_j) where p_i is a tuple containing a
            primer_search.PrimerResult object, and likewise for p_j.
            The start position of p_j is > the start position of p_i.
        """
        primers = list(self.ps.find_primers(
            max_at_site=self.max_primers_at_site))
        for i in range(len(primers) - 1):
            p1 = primers[i]
            for j in range(i + 1, len(primers)):
                p2 = primers[j]
                yield (p1, p2)

    def _find_mutated_activity(self, targets):
        """Find the activities of the guides after mutation

        For each target, find the activity of the best of the guide
        set across each potential starting location after mutation,
        averaged across the sequences.

        Args:
            targets: list of tuples (obj_value, target) where target is a tuple
                ((p1, p2), (guides_stats, guides))
        Returns:
            list of mutated activities, ordered by target's ordering.

        """
        mutated_activities = [None] * len(targets)
        for i, (obj_value, target) in enumerate(targets):
            ((p1, p2), (guides_stats, guides)) = target
            mutated_activities[i] = max(max(np.average(
                        self.gs.aln.compute_activity(
                            start_pos,
                            guide,
                            self.gs.predictor,
                            self.mutator))
                    for start_pos in self.gs._selected_guide_positions[guide])
                for guide in guides)
        return mutated_activities

    def find_targets(self, best_n=10, no_overlap='amplicon'):
        """Find targets across an alignment.

        Args:
            best_n: only store and output the best_n targets according
                to the objective
            no_overlap: if 'amplicon', do not allow targets in the output
                whose amplicons (target range) overlap; if 'primer', do not
                allow targets in the output if both primers overlap.
                When not 'none', if a target overlaps with a target already
                in the output, this *replaces* the overlapping one with
                the new one if the new one has a better objective value.
                When 'none', many targets in the best_n may be very similar

        Returns:
            list of tuples (obj_value, target) where target is a tuple
            ((p1, p2), (guides_stats, guides)); (p1, p2) is a pair
            of PrimerResult objects, and guides is a list of guides that
            meet the constraints within the window bound by (p1, p2)
        """

        # Store a heap containing the best_n primer/guide targets;
        # store each as a tuple ((+/-)obj_value, push_id, target)
        # The heap is a min heap, so popped values are the ones with
        # the smallest value in the first element
        #   - When self.obj_type is 'min' we store -obj_value in the
        #     first element so that the one with the highest objective
        #     value (worst) is popped.
        #   - When self.obj_type is 'max' we store +obj_value in
        #     the first element so that the one with the smallest
        #     objective value (worst) is popped
        # In both cases, target_heap[0] refers to the target with the
        # worst objective value
        # We also store push_id (a counter on the push) to break ties
        # when costs are equivalent (choosing the one pushed first, i.e.,
        # the lower push_id); otherwise, there will be an error on
        # ties if target is not comparable
        target_heap = []
        push_id_counter = itertools.count()

        assert self.obj_type in ['min', 'max']
        assert no_overlap in ['amplicon', 'primer', 'none']
        def obj_value(i):
            # Return objective value of the i'th element
            if self.obj_type == 'min':
                return -1 * target_heap[i][0]
            elif self.obj_type == 'max':
                return target_heap[i][0]

        def obj_value_is_better(new, old):
            # Compare new objective value to old, and return True iff
            # new is better
            if self.obj_type == 'min':
                if new < old:
                    return True
            elif self.obj_type == 'max':
                if new > old:
                    return True
            return False

        # Determine a best-case objective value from the guide search
        # This is useful for pruning the search
        best_guide_obj_value = self.gs.best_obj_value()

        num_primer_pairs = 0
        num_suitable_primer_pairs = 0
        last_window_start = -1
        for p1, p2 in self._find_primer_pairs():
            if self.halt_early and len(target_heap) >= best_n:
                # Enough targets were found
                # Stop early without completing the search
                break

            num_primer_pairs += 1

            target_length = p2.start + p2.primer_length - p1.start
            if (self.max_target_length is not None and
                    target_length > self.max_target_length):
                # This is longer than allowed, so skip it
                continue

            # Determine a window between the two primers
            window_start = p1.start + p1.primer_length
            window_end = p2.start

            # To be a suitable window, p2 must start after p1 ends
            if window_end <= window_start:
                continue
            window_length = window_end - window_start

            # To consider this window, a guide must fit within it
            if window_length < self.gs.guide_length:
                continue

            # If here, the window passed basic checks
            num_suitable_primer_pairs += 1

            # Since window_start increases monotonically, we no
            # longer need to memoize guide covers between the last
            # and current start
            assert window_start >= last_window_start
            for pos in range(last_window_start, window_start):
                self.gs._cleanup_memoized_guides(pos)
            last_window_start = window_start

            # Calculate a cost of the primers
            p1_num = p1.num_primers
            p2_num = p2.num_primers
            cost_primers = self.obj_weight_primers * (p1_num + p2_num)

            # Calculate a cost of the window length
            cost_window = self.obj_weight_length * math.log2(window_length)

            # Calculate a best-case objective value for this target,
            # which can be done assuming a best-case for the guide search
            # This is useful for pruning the search
            if self.obj_type == 'min':
                best_possible_obj_value = (best_guide_obj_value +
                        cost_primers + cost_window)
            elif self.obj_type == 'max':
                best_possible_obj_value = (best_guide_obj_value -
                        cost_primers - cost_window)

            # Check if we should bother trying to find guides in this window
            if len(target_heap) >= best_n:
                curr_worst_obj_value = obj_value(0)
                if self.obj_type == 'min':
                    if best_possible_obj_value >= curr_worst_obj_value:
                        # The objective value is already >= than all in the
                        # current best_n, and will only stay the same or get
                        # bigger after adding the term for guides, so there is
                        # no reason to continue considering this window
                        continue
                elif self.obj_type == 'max':
                    if best_possible_obj_value <= curr_worst_obj_value:
                        # The objective value is <= all in the current best_n
                        # even in the best-case outcome for the guide search,
                        # and it will only stay the same or be smaller after
                        # adding the term for guides, so there is no reason to
                        # continue considering this window
                        continue

            # If targets should not overlap and this overlaps a target
            # already in target_heap, check if we should bother trying to
            # find guides in this window
            # Note that this strategy does have the possibility of leading
            # to suboptimal solutions. Effectively, successive windows
            # can repeatedly remove other overlapping ones that might
            # otherwise be in the best_n. For example, consider targets
            # A, B, and C where A overlaps B and B overlaps C but A
            # does not overlap C. And say the true ranking is
            # A<B<C, but their objective values are close and better than
            # any other possible target option. Adding B would remove
            # A, and then adding C would remove B. So we are left with
            # only C from this set of targets, but ideally we would
            # like to have both A and C. One way around this might be
            # to keep a heap with >best_n options and only prune for
            # diverse solutions (non-overlap) later on.
            overlapping_i = []
            if no_overlap == 'amplicon':
                # Check if any targets already in target_heap overlap
                # with this target
                this_start = p1.start
                this_end = this_start + target_length
                for i in range(len(target_heap)):
                    _, _, target_i = target_heap[i]
                    (p1_i, p2_i), _ = target_i
                    i_start = p1_i.start
                    i_end = p2_i.start + p2_i.primer_length
                    if (this_start < i_end) and (i_start < this_end):
                        # Replace target_i
                        overlapping_i += [i]
            if no_overlap == 'primer':
                for i in range(len(target_heap)):
                    _, _, target_i = target_heap[i]
                    (p1_i, p2_i), _ = target_i
                    if p1.overlaps(p1_i) and p2.overlaps(p2_i):
                        # Replace target_i
                        overlapping_i += [i]
            if overlapping_i:
                # target overlaps with one or more entries already in
                # target_heap
                could_replace = False
                for i in overlapping_i:
                    if obj_value_is_better(best_possible_obj_value,
                            obj_value(i)):
                        could_replace = True
                        break
                if could_replace is False:
                    # We will try to replace an entry in overlapping_i (i) with
                    # a new entry, but the objective value is such that, even
                    # in the best-case, it will not replace i. So there is
                    # no reason for considering this window (it will never
                    # replace i)
                    continue

            logger.info(("Found window [%d, %d) bound by primers that could "
                "be in the best %d targets; looking for guides within this; "
                "heap has %d valid targets currently"),
                window_start, window_end, best_n, len(target_heap))

            # Determine what set of sequences the guides should account for
            if self.only_account_for_amplified_seqs:
                # Find the sequences that are bound by some primer in p1 AND
                # some primer in p2
                # Note that this is only implemented when self.gs is an
                # instance of GuideSearcherMinimizeGuides (i.e., not for
                # MaximizeActivity -- mostly for technical implementation
                # reasons)
                p1_bound_seqs = self.ps.seqs_bound_by_primers(p1.primers_in_cover)
                p2_bound_seqs = self.ps.seqs_bound_by_primers(p2.primers_in_cover)
                primer_bound_seqs = p1_bound_seqs & p2_bound_seqs
                guide_seqs_to_consider = primer_bound_seqs
            else:
                # Design across the entire collection of sequences
                # This may lead to more guides being designed than if only
                # designing across primer_bound_seqs (below) because more
                # sequences must be considered in the design
                # But this can improve runtime because the seqs_to_consider
                # used by self.gs in the design will be more constant
                # across different windows, which will enable it to better
                # take advantage of memoization (in self.gs._memoized_guides)
                # during the design
                guide_seqs_to_consider = None

            # Find guides in the window
            if isinstance(self.gs, guide_search.GuideSearcherMinimizeGuides):
                extra_args = {'only_consider': guide_seqs_to_consider}
            else:
                extra_args = {}
            try:
                guides = self.gs._find_guides_in_window(
                    window_start, window_end, **extra_args)
            except guide_search.CannotAchieveDesiredCoverageError:
                # No more suitable guides; skip this window
                continue
            except guide_search.CannotFindAnyGuidesError:
                # No suitable guides in this window; skip it
                continue

            if len(guides) == 0:
                # This can happen, for example, if primer_bound_seqs is
                # empty; then no guides are required
                # Skip this window
                continue

            # Compute activities across target sequences, and expected, median,
            # and 5th percentile of activities
            if self.gs.predictor is not None:
                activities = self.gs.guide_set_activities(window_start,
                        window_end, guides)
                guides_activity_expected = self.gs.guide_set_activities_expected_value(
                        window_start, window_end, guides,
                        activities=activities)
                guides_activity_median, guides_activity_5thpctile = \
                        self.gs.guide_set_activities_percentile(
                                window_start, window_end, guides, [50, 5],
                                activities=activities)
            else:
                # There is no predictor to predict activities
                # This should only be the case if self.obj_type is 'min',
                # and may not necessarily be the case if self.obj_type is
                # 'min'
                activities = None
                guides_activity_expected = math.nan
                guides_activity_median, guides_activity_5thpctile = \
                        math.nan, math.nan
            # Calculate fraction of sequences bound by the guides
            if isinstance(self.gs, guide_search.GuideSearcherMinimizeGuides):
                guides_frac_bound = self.gs.total_frac_bound_by_guides(guides)
            elif isinstance(self.gs, guide_search.GuideSearcherMaximizeActivity):
                guides_frac_bound = self.gs.total_frac_bound_by_guides(
                        window_start, window_end, guides,
                        activities=activities)
            guides_stats = (guides_frac_bound, guides_activity_expected,
                    guides_activity_median, guides_activity_5thpctile)

            # Calculate a total objective value
            if self.obj_type == 'min':
                obj_value_total = (self.gs.obj_value(guides) +
                        cost_primers + cost_window)
                obj_value_to_add = -1 * obj_value_total
            elif self.obj_type == 'max':
                gs_obj_value = self.gs.obj_value(window_start, window_end,
                        guides, activities=activities)
                obj_value_total = (gs_obj_value -
                        cost_primers - cost_window)
                obj_value_to_add = obj_value_total

            # Add target to the heap (but only keep it if there are not
            # yet best_n targets or it has one of the best_n best objective
            # values)
            target = ((p1, p2), (guides_stats, guides))
            entry = (obj_value_to_add, next(push_id_counter), target)
            if overlapping_i:
                # target overlaps with an entry already in target_heap;
                # consider replacing that entry with new entry
                if len(overlapping_i) == 1:
                    # Almost all cases will satisfy this: the new entry
                    # overlaps just one existing entry; optimize for this case
                    i = overlapping_i[0]
                    if obj_value_is_better(obj_value_total, obj_value(i)):
                        # Replace i with entry
                        target_heap[i] = entry
                        heapq.heapify(target_heap)
                else:
                    # For some edge cases, the new entry overlaps with
                    # multiple existing entries
                    # Check if the new entry has a sufficiently low cost
                    # to justify removing any existing overlapping entries,
                    # where 'sufficiently low cost' means a lower cost than
                    # all of the overlapping targets
                    obj_value_is_sufficiently_good = True
                    for i in overlapping_i:
                        if obj_value_is_better(obj_value(i), obj_value_total):
                            obj_value_is_sufficiently_good = False
                            break
                    if obj_value_is_sufficiently_good:
                        # Remove all entries that overlap the new entry,
                        # then add the new entry
                        for i in sorted(overlapping_i, reverse=True):
                            del target_heap[i]
                        target_heap.append(entry)
                        heapq.heapify(target_heap)
                        # Note that this can lead to cases in which the size
                        # of the output (target_heap) is < best_n, or in which
                        # there is an entry with poor objective value, because
                        # in certain edge cases we might finish the search
                        # having removed >1 entry but only replacing it with
                        # one, and if the search continues for a short time
                        # after, len(target_heap) < best_n could result in
                        # some poor entries being placed into the heap
            elif len(target_heap) < best_n:
                # target_heap is not yet full, so push entry to it
                heapq.heappush(target_heap, entry)
            else:
                # target_heap is full; consider replacing the worst entry
                # with the new entry
                curr_worst_obj_value = obj_value(0)
                if obj_value_is_better(obj_value_total, curr_worst_obj_value):
                    # Push the entry into target_heap, and pop out the
                    # worst one
                    heapq.heappushpop(target_heap, entry)

        if len(target_heap) == 0:
            logger.warning(("Zero targets were found. The number of total "
                "primer pairs found was %d and the number of them that "
                "were suitable (passing basic criteria, e.g., on length) "
                "was %d"), num_primer_pairs, num_suitable_primer_pairs)

        # Invert the objective values if they were inverted (target_heap had
        # been storing the negative of each objective value for minimization),
        # toss push_id, and sort by objective value
        # In particular, sort by a 'sort_tuple' whose first element is
        # objective_value and then the target endpoints; it has the target
        # endpoints to break ties in objective value
        if self.obj_type == 'min':
            r = [(-obj_value, target) for obj_value, push_id, target in target_heap]
            r_with_sort_tuple = []
            for (obj_value, target) in r:
                ((p1, p2), (guides_stats, guides)) = target
                target_start = p1.start
                # Sort all fields of sort_tuple in ascending order
                sort_tuple = (obj_value, target_start)
                r_with_sort_tuple += [(sort_tuple, obj_value, target)]
            r_with_sort_tuple = sorted(r_with_sort_tuple, key=lambda x: x[0])
        elif self.obj_type == 'max':
            r = [(obj_value, target) for obj_value, push_id, target in target_heap]
            r_with_sort_tuple = []
            for (obj_value, target) in r:
                ((p1, p2), (guides_stats, guides)) = target
                target_start = p1.start
                # Sort obj_value in descending order, and target_start in
                # ascending order
                sort_tuple = (-1 * obj_value, target_start)
                r_with_sort_tuple += [(sort_tuple, obj_value, target)]
            r_with_sort_tuple = sorted(r_with_sort_tuple, key=lambda x: x[0])
        r = [(obj_value, target) for sort_tuple, obj_value, target in r_with_sort_tuple]

        # Shift obj_value
        r = [(obj_value + self.obj_value_shift, target)
                for obj_value, target in r]

        return r

    def find_and_write_targets(self, out_fn, best_n=10, no_overlap='amplicon'):
        """Find targets across an alignment, and write them to a file.

        This writes a table of the targets to a file, in which each
        row corresponds to a target in the genome (primer and guide
        sets).

        Args:
            out_fn: output TSV file to write targets
            best_n: only store and output the best_n targets according to
                objective value
            no_overlap: if 'amplicon', do not allow targets in the output
                whose amplicons (target range) overlap; if 'primer', do not
                allow targets in the output if both primers overlap.
                When not 'none', if a target overlaps with a target already
                in the output, this *replaces* the overlapping one with
                the new one if the new one has a better objective value.
                When 'none', many targets in the best_n may be very similar
        """
        targets = self.find_targets(best_n=best_n, no_overlap=no_overlap)
        if self.mutator:
            mutated_activity = self._find_mutated_activity(targets)

        with open(out_fn, 'w') as outf:
            # Write a header
            headers = ['objective-value', 'target-start', 'target-end',
                'target-length',
                'left-primer-start', 'left-primer-num-primers',
                'left-primer-frac-bound', 'left-primer-target-sequences',
                'right-primer-start', 'right-primer-num-primers',
                'right-primer-frac-bound', 'right-primer-target-sequences',
                'num-guides', 'total-frac-bound-by-guides',
                'guide-set-expected-activity',
                'guide-set-median-activity', 'guide-set-5th-pctile-activity',
                'guide-expected-activities',
                'guide-target-sequences', 'guide-target-sequence-positions']
            if self.mutator:
                headers.append('guide-set-5th-pctile-mutated-activity')

            outf.write('\t'.join(headers) + '\n')

            for i, (obj_value, target) in enumerate(targets):
                ((p1, p2), (guides_stats, guides)) = target

                # Break out guides_stats
                guides_frac_bound, guides_activity_expected, guides_activity_median, guides_activity_5thpctile = guides_stats

                # Determine the target endpoints
                target_start = p1.start
                target_end = p2.start + p2.primer_length
                target_length = target_end - target_start

                # Construct string of primers and guides
                p1_seqs_sorted = sorted(list(p1.primers_in_cover))
                p1_seqs_str = ' '.join(p1_seqs_sorted)
                p2_seqs_sorted = sorted(list(p2.primers_in_cover))
                p2_seqs_str = ' '.join(p2_seqs_sorted)
                guides_seqs_sorted = sorted(list(guides))
                guides_seqs_str = ' '.join(guides_seqs_sorted)

                # Find positions of the guides
                guides_positions = [self.gs._selected_guide_positions[gd_seq]
                                    for gd_seq in guides_seqs_sorted]
                guides_positions_str = ' '.join(str(p) for p in guides_positions)

                # Find expected activity for each guide
                window_start = p1.start + p1.primer_length
                window_end = p2.start
                if self.gs.predictor is not None:
                    expected_activities_per_guide = \
                            [self.gs.guide_activities_expected_value(
                                window_start, window_end, gd_seq)
                                for gd_seq in guides_seqs_sorted]
                else:
                    # There is no predictor to predict activities
                    # This should only be the case if self.obj_type is 'min',
                    # and may not necessarily be the case if self.obj_type is
                    # 'min'
                    expected_activities_per_guide = \
                            [math.nan for gd_seq in guides_seqs_sorted]
                expected_activities_per_guide_str = ' '.join(
                        str(a) for a in expected_activities_per_guide)

                line = [obj_value, target_start, target_end, target_length,
                    p1.start, p1.num_primers, p1.frac_bound, p1_seqs_str,
                    p2.start, p2.num_primers, p2.frac_bound, p2_seqs_str,
                    len(guides), guides_frac_bound, guides_activity_expected,
                    guides_activity_median, guides_activity_5thpctile,
                    expected_activities_per_guide_str,
                    guides_seqs_str, guides_positions_str]
                if self.mutator:
                    line.append(mutated_activity[i])

                outf.write('\t'.join([str(x) for x in line]) + '\n')


class DesignTarget:
    """Store information on a design of a single target.
    """

    def __init__(self, target_start, target_end, guide_seqs,
            left_primer_seqs, right_primer_seqs, obj_value):
        self.target_start = target_start
        self.target_end = target_end
        self.guide_seqs = tuple(sorted(guide_seqs))
        self.left_primer_seqs = tuple(sorted(left_primer_seqs))
        self.right_primer_seqs = tuple(sorted(right_primer_seqs))
        self.obj_value = obj_value

    @staticmethod
    def read_design_targets(fn, num_targets=None):
        """Read a collection of targets from a file.

        Args:
            fn: path to a TSV file giving targets
            num_targets: only construct a Design from the top num_targets
                targets, as ordered by cost (if None, use all)

        Returns:
            list of DesignTarget objects
        """
        if not os.path.isfile(fn):
            return None

        rows = []
        has_objective_value = False
        with open(fn) as f:
            col_names = {}
            for i, line in enumerate(f):
                line = line.rstrip()
                ls = line.split('\t')
                if i == 0:
                    # Parse header
                    for j in range(len(ls)):
                        col_names[j] = ls[j]
                else:
                    # Read each column as a variable
                    cols = {}
                    for j in range(len(ls)):
                        cols[col_names[j]] = ls[j]
                    if has_objective_value:
                        rows += [(cols['objective-value'], cols['target-start'],
                                 cols['target-end'], cols)]
                    else:
                        rows += [cols]

        # Pull out the best N targets, assuming rows are already sorted
        # as desired
        if num_targets != None:
            if has_objective_value is False:
                raise Exception(("Cannot pull out best targets; objective "
                    "value is not given"))
            if len(rows) < num_targets:
                raise Exception(("The number of rows in a design (%d) is fewer "
                    "than the number of targets to read (%d)") %
                    (len(rows), num_targets))
            rows = rows[:num_targets]

        targets = []
        for row in rows:
            if has_objective_value:
                _, _, _, cols = row
                obj_value = float(cols['objective-value'])
            else:
                cols = row
                obj_value = None
            targets += [DesignTarget(
                int(cols['target-start']),
                int(cols['target-end']),
                cols['guide-target-sequences'].split(' '),
                cols['left-primer-target-sequences'].split(' '),
                cols['right-primer-target-sequences'].split(' '),
                obj_value
            )]

        return targets



"""Use GTR to model distribution of future viral sequences
"""

import numpy as np
from adapt import alignment
from adapt.utils import predict_activity
from scipy.linalg import expm

__author__ = 'David K. Yang <yangd5153@gmail.com>, Priya P. Pillai <ppillai@broadinstitute.org>'


class GTRSubstitutionMutator:
    """Use the GTR Substitution model to mutate viral sequences
    """
    def __init__(self, aln,
                 rAC, rAG, rAT, rCG, rCT, rGT,
                 mu, t, n):
        """
        Args:
            aln: An alignment of sequences to use to determine base percentages
            rAC: Relative rate of conversion between A and C
            rAG: Relative rate of conversion between A and G
            rAT: Relative rate of conversion between A and T
            rCG: Relative rate of conversion between C and G
            rCT: Relative rate of conversion between C and T
            rGT: Relative rate of conversion between G and T
            mu: Overall rate of substitutions per site per year
            t: Years to simulate substitutions over
            n: Number of sequences to simulate mutations over
        """
        base_percentages = aln.base_percentages()
        self.piA = base_percentages['A']
        self.piC = base_percentages['C']
        self.piG = base_percentages['G']
        self.piT = base_percentages['T']
        self.rAC = rAC
        self.rAG = rAG
        self.rAT = rAT
        self.rCG = rCG
        self.rCT = rCT
        self.rGT = rGT
        self.mu = mu
        self.t = t
        self.n = n

        self.Q = self._construct_rateQ()
        self.P = self._construct_P()

    def _construct_rateQ(self):
        """Compute transition rate matrix

        Computes transition rate matrix Q, given base frequencies and
        transition rates under GTR model (Tavaré 1986).

        The transition rate matrix defines the rate each base is mutated to
        each other base. The rows indicate the starting base; the columns
        indicate the final base. Bases are ordered A, C, G, T. Diagonal
        elements are set such that the row sums to 0.
        """

        beta = 1 / (2 * (
            self.piA * (self.rAC*self.piC +
                        self.rAG*self.piG +
                        self.rAT*self.piT) +
            self.piC * (self.rCG*self.piG +
                        self.rCT*self.piT) +
            self.piG * (self.rGT*self.piT)
        ))
        Q = np.array([
                [-(self.rAC*self.piC + self.rAG*self.piG + self.rAT*self.piT),
                 self.rAC*self.piC,
                 self.rAG*self.piG,
                 self.rAT*self.piT],
                [self.rAC*self.piA,
                 -(self.rAC*self.piA + self.rCG*self.piG + self.rCT*self.piT),
                 self.rCG*self.piG,
                 self.rCT*self.piT],
                [self.rAG*self.piA,
                 self.rCG*self.piC,
                 -(self.rAG*self.piA + self.rCG*self.piC + self.rGT*self.piT),
                 self.rGT*self.piT],
                [self.rAT*self.piA,
                 self.rCT*self.piC,
                 self.rGT*self.piG,
                 -(self.rAT*self.piA + self.rCT*self.piC + self.rGT*self.piG)]
            ])

        return beta * Q

    def _construct_P(self):
        """Compute transition probability matrix

        Computes transition probability matrix P from rate matrix Q,
        substitution rate m, and time t

        The transition probablility matrix defines the likelihood each base is
        mutated to each other base. The rows indicate the starting base; the
        columns indicate the final base. Bases are ordered A, C, G, T.
        """

        P = expm(self.Q * self.mu * self.t)
        # Normalize so each row sums to 1.
        # Matrix exponentiation should already do this in principle
        row_sums = P.sum(axis=1)
        normalized_matrix = P / row_sums[:, np.newaxis]
        return normalized_matrix

    def _seq_to_encoding(self, seq):
        """Encode string sequence into a nucleotide index list

        Map 'A' to 0, 'C' to 1, 'G' to 2, and 'T' to 3
        (e.g. "ACGT" returns [0, 1, 2, 3])

        Args:
            str: nucleotide sequence

        Returns:
            list(int): list of nucleotide indexs.
        """

        base_key = "ACGT"
        seq_as_list = list(seq)
        return [base_key.index(base) for base in seq_as_list]

    def compute_sequence_probability(self, wild_seq, mut_seq):
        """Compute probability of wild_seq mutating into mut_seq

        Under transition probability matrix P, finds probability of wild_seq
        mutating into mut_seq

        Args:
            wild_seq: reference sequence
            mut_seq: mutated sequence to compute probability for

        Returns:
            float: probability of mutated sequence
        """

        wild_seq_encoded = self._seq_to_encoding(wild_seq)
        mut_seq_encoded = self._seq_to_encoding(mut_seq)

        prob = 1.0
        for a, b in zip(wild_seq_encoded, mut_seq_encoded):
            prob = prob * self.P[a][b]

        return prob

    def mutate(self, wild_seq):
        """Sample future sequences under transition probability matrix P

        Args:
            wild_seq: reference sequence

        Returns:
            list(str): list of samples from the distribution of future
                sequences
        """

        wild_seq_encoded = self._seq_to_encoding(wild_seq)
        sampled_seq_matrix = []

        for res in wild_seq_encoded:
            samples = np.random.choice(["A", "C", "G", "T"], self.n,
                                       p=self.P[res])
            sampled_seq_matrix.append(samples)

        sampled_seqs_list = np.array(sampled_seq_matrix).transpose()
        return ["".join(seq) for seq in sampled_seqs_list]

    def compute_mutated_activity(self, predictor, target_seq,
                                 guide_seq, start=0):
        """Calculate the activity of the guide after mutating the target

        Args:
            predictor: an adapt.utils.predict_activity Predictor object
            target_seq: string of what sequence the guide targets. Includes
                    context if predictor requires it
            guide_seq: string of what the guide is
            start: int, start position for the guide sequence in the alignment.
                    Required to use predictor memoization, if it exists

        Returns:
            The 5th percentile of activity values from simulated mutations
        """
        mutated_target_seqs = self.mutate(target_seq)
        if isinstance(predictor, predict_activity.SimpleBinaryPredictor):
            mutated_aln = alignment.Alignment.from_list_of_seqs(
                mutated_target_seqs)
            left_context = 0
            if predictor.required_flanking_seqs[0]:
                left_context = len(predictor.required_flanking_seqs[0])
            _, activity = predictor.compute_activity(left_context, guide_seq,
                                                     mutated_aln, percentiles=5)
        else:
            pairs_to_eval = []
            for mutated_target_seq in mutated_target_seqs:
                pair = (mutated_target_seq, guide_seq)
                pairs_to_eval.append(pair)
            _, activity = predictor.compute_activity(start, pairs_to_eval,
                                                     percentiles=5)
        return activity



"""Tests for index_compress module.
"""

import unittest

from adapt.utils import index_compress

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestIndexCompress(unittest.TestCase):
    """Tests functions for compressing/de-compressing integers.
    """

    def run_compress_and_decompress(self, idx):
        compressed = index_compress.compress_mostly_contiguous(idx)
        decompressed = index_compress.decompress_ranges(compressed)
        self.assertEqual(decompressed, idx)

    def test_contiguous(self):
        self.run_compress_and_decompress({0, 1, 2, 3})
        self.run_compress_and_decompress({1, 2, 3})

    def test_uncontinguous(self):
        self.run_compress_and_decompress({0, 1, 2, 4, 5})
        self.run_compress_and_decompress({0, 1, 2, 4, 5, 7})
        self.run_compress_and_decompress({0, 2, 4, 5, 7, 8, 10})
        self.run_compress_and_decompress({0, 2, 4, 5, 7, 8, 10, 12})

    def test_duplicate(self):
        self.run_compress_and_decompress({0, 1, 2, 2, 3})
        self.run_compress_and_decompress({0, 1, 2, 2, 3, 5, 5})


class TestIndexCompressForDicts(unittest.TestCase):
    """Tests functions for compressing/de-compressing dicts.
    """

    def run_compress_and_decompress(self, d):
        compressed = index_compress.compress_mostly_contiguous_keys_with_identical_vals(d)
        decompressed = index_compress.decompress_ranges_for_dict(compressed)
        self.assertEqual(decompressed, d)

    def test_contiguous(self):
        self.run_compress_and_decompress({0:300, 1:100, 2:100, 3:200})
        self.run_compress_and_decompress({1:100, 2:200, 3:200})
        self.run_compress_and_decompress({1:100, 2:200, 3:100})

    def test_uncontinguous(self):
        self.run_compress_and_decompress({0:100, 1:100, 2:100, 4:100, 5:100})
        self.run_compress_and_decompress({0:200, 1:100, 2:200, 4:200, 5:200,
            7:200})
        self.run_compress_and_decompress({0:200, 2:200, 4:300, 5:300, 7:400,
            8:500, 10:500})
        self.run_compress_and_decompress({0:200, 2:300, 4:400, 5:500, 7:700,
            8:800, 10:1000, 12:1000})

